import numbers

import numpy as np

from gpaw.response import ResponseGroundStateAdapter, ResponseContext, timer
from gpaw.response.pw_parallelization import block_partition
from gpaw.response.symmetry import KPointFinder
from gpaw.utilities.blas import mmm


class KPoint:
    def __init__(self, s, K, n1, n2, blocksize, na, nb,
                 ut_nR, eps_n, f_n, P_ani, k_c):
        self.s = s    # spin index
        self.K = K    # BZ k-point index
        self.n1 = n1  # first band
        self.n2 = n2  # first band not included
        self.blocksize = blocksize
        self.na = na  # first band of block
        self.nb = nb  # first band of block not included
        self.ut_nR = ut_nR      # periodic part of wave functions in real-space
        self.eps_n = eps_n      # eigenvalues
        self.f_n = f_n          # occupation numbers
        self.P_ani = P_ani      # PAW projections
        self.k_c = k_c  # k-point coordinates


class PairDistribution:
    def __init__(self, pair, mysKn1n2):
        self.pair = pair
        self.mysKn1n2 = mysKn1n2
        self.mykpts = [self.pair.get_k_point(s, K, n1, n2)
                       for s, K, n1, n2 in self.mysKn1n2]

    def kpt_pairs_by_q(self, q_c, m1, m2):
        pair = self.pair
        mykpts = self.mykpts
        for u, kpt1 in enumerate(mykpts):
            progress = u / len(mykpts)
            K2 = pair.gs.kd.find_k_plus_q(q_c, [kpt1.K])[0]
            kpt2 = pair.get_k_point(kpt1.s, K2, m1, m2, block=True)

            yield progress, kpt1, kpt2


class KPointPair:
    """This class defines the kpoint-pair container object.

    Used for calculating pair quantities it contains two kpoints,
    and an associated set of Fourier components."""
    def __init__(self, kpt1, kpt2, Q_G):
        self.kpt1 = kpt1
        self.kpt2 = kpt2
        self.Q_G = Q_G

    def get_transition_energies(self, n_n, m_m):
        """Return the energy difference for specified bands."""
        n_n = np.array(n_n)
        m_m = np.array(m_m)
        kpt1 = self.kpt1
        kpt2 = self.kpt2
        deps_nm = (kpt1.eps_n[n_n - self.kpt1.n1][:, np.newaxis] -
                   kpt2.eps_n[m_m - self.kpt2.n1])
        return deps_nm

    def get_occupation_differences(self, n_n, m_m):
        """Get difference in occupation factor between specified bands."""
        n_n = np.array(n_n)
        m_m = np.array(m_m)
        kpt1 = self.kpt1
        kpt2 = self.kpt2
        df_nm = (kpt1.f_n[n_n - self.kpt1.n1][:, np.newaxis] -
                 kpt2.f_n[m_m - self.kpt2.n1])
        return df_nm


class KPointPairFactory:
    def __init__(self, gs, context, *, nblocks=1):
        self.gs = gs
        self.context = context

        assert self.gs.kd.symmetry.symmorphic

        self.blockcomm, self.kncomm = block_partition(self.context.comm,
                                                      nblocks)
        self.nblocks = nblocks

        self.kptfinder = KPointFinder(self.gs.kd.bzk_kc)
        self.context.print('Number of blocks:', nblocks)

    def find_kpoint(self, k_c):
        return self.kptfinder.find(k_c)

    def distribute_k_points_and_bands(self, band1, band2, kpts=None):
        """Distribute spins, k-points and bands.

        The attribute self.mysKn1n2 will be set to a list of (s, K, n1, n2)
        tuples that this process handles.
        """

        gs = self.gs

        if kpts is None:
            kpts = np.arange(gs.kd.nbzkpts)

        # nbands is the number of bands for each spin/k-point combination.
        nbands = band2 - band1
        size = self.kncomm.size
        rank = self.kncomm.rank
        ns = gs.nspins
        nk = len(kpts)
        n = (ns * nk * nbands + size - 1) // size
        i1 = min(rank * n, ns * nk * nbands)
        i2 = min(i1 + n, ns * nk * nbands)

        mysKn1n2 = []
        i = 0
        for s in range(ns):
            for K in kpts:
                n1 = min(max(0, i1 - i), nbands)
                n2 = min(max(0, i2 - i), nbands)
                if n1 != n2:
                    mysKn1n2.append((s, K, n1 + band1, n2 + band1))
                i += nbands

        p = self.context.print
        p('BZ k-points:', gs.kd, flush=False)
        p('Distributing spins, k-points and bands (%d x %d x %d)' %
          (ns, nk, nbands), 'over %d process%s' %
          (self.kncomm.size, ['es', ''][self.kncomm.size == 1]),
          flush=False)
        p('Number of blocks:', self.blockcomm.size)

        return PairDistribution(self, mysKn1n2)

    @timer('Get a k-point')
    def get_k_point(self, s, k_c, n1, n2, block=False):
        """Return wave functions for a specific k-point and spin.

        s: int
            Spin index (0 or 1).
        K: int
            BZ k-point index.
        n1, n2: int
            Range of bands to include.
        """

        assert n1 <= n2

        gs = self.gs
        kd = gs.kd

        # Parse kpoint: is k_c an index or a vector
        if not isinstance(k_c, numbers.Integral):
            K = self.kptfinder.find(k_c)
        else:
            # Fall back to index
            K = k_c

        if block:
            nblocks = self.blockcomm.size
            rank = self.blockcomm.rank
        else:
            nblocks = 1
            rank = 0

        blocksize = (n2 - n1 + nblocks - 1) // nblocks
        na = min(n1 + rank * blocksize, n2)
        nb = min(na + blocksize, n2)

        ik = kd.bz2ibz_k[K]
        assert kd.comm.size == 1
        kpt = gs.kpt_qs[ik][s]

        assert n2 <= len(kpt.eps_n), \
            'Increase GS-nbands or decrease chi0-nbands!'
        eps_n = kpt.eps_n[n1:n2]
        f_n = kpt.f_n[n1:n2] / kpt.weight

        k_c = self.gs.ibz2bz[K].map_kpoint()

        with self.context.timer('load wfs'):
            psit_nG = kpt.psit_nG
            ut_nR = gs.gd.empty(nb - na, gs.dtype)
            for n in range(na, nb):
                ut_nR[n - na] = self.gs.ibz2bz[K].map_pseudo_wave(
                    gs.pd.ifft(psit_nG[n], ik))

        with self.context.timer('Load projections'):
            if nb - na > 0:
                proj = kpt.projections.new(nbands=nb - na, bcomm=None)
                proj.array[:] = kpt.projections.array[na:nb]
                proj = self.gs.ibz2bz[K].map_projections(proj)
                P_ani = [P_ni for _, P_ni in proj.items()]
            else:
                P_ani = []

        return KPoint(s, K, n1, n2, blocksize, na, nb,
                      ut_nR, eps_n, f_n, P_ani, k_c)

    @timer('Get kpoint pair')
    def get_kpoint_pair(self, qpd, s, Kork_c, n1, n2, m1, m2, block=False):
        assert m1 <= m2
        assert n1 <= n2

        if isinstance(Kork_c, int):
            # If k_c is an integer then it refers to
            # the index of the kpoint in the BZ
            k_c = self.gs.kd.bzk_kc[Kork_c]
        else:
            k_c = Kork_c

        q_c = qpd.q_c
        with self.context.timer('get k-points'):
            kpt1 = self.get_k_point(s, k_c, n1, n2)
            # K2 = wfs.kd.find_k_plus_q(q_c, [kpt1.K])[0]
            kpt2 = self.get_k_point(s, k_c + q_c, m1, m2, block=block)

        with self.context.timer('fft indices'):
            Q_G = phase_shifted_fft_indices(kpt1.k_c, kpt2.k_c, qpd)

        return KPointPair(kpt1, kpt2, Q_G)

    def pair_calculator(self):
        # We have decoupled the actual pair density calculator
        # from the kpoint factory, but it's still handy to
        # keep this shortcut -- for now.
        return ActualPairDensityCalculator(self)


class ActualPairDensityCalculator:
    def __init__(self, pair):
        self.context = pair.context
        self.blockcomm = pair.blockcomm
        self.ut_sKnvR = None  # gradient of wave functions for optical limit
        self.gs = pair.gs

    def get_optical_pair_density(self, qpd, kptpair, n_n, m_m, *,
                                 pawcorr, block=False):
        """Get the full optical pair density, including the optical limit head
        for q=0."""
        tmp_nmG = self.get_pair_density(qpd, kptpair, n_n, m_m,
                                        pawcorr=pawcorr, block=block)

        nG = qpd.ngmax
        # P = (x, y, z, G1, G2, ...)
        n_nmP = np.empty((len(n_n), len(m_m), nG + 2), dtype=tmp_nmG.dtype)
        n_nmP[:, :, 3:] = tmp_nmG[:, :, 1:]
        n_nmv = self.get_optical_pair_density_head(qpd, kptpair, n_n, m_m,
                                                   block=block)
        n_nmP[:, :, :3] = n_nmv

        return n_nmP

    @timer('get_pair_density')
    def get_pair_density(self, qpd, kptpair, n_n, m_m, *,
                         pawcorr, block=False):
        """Get pair density for a kpoint pair."""
        cpd = self.calculate_pair_density

        kpt1 = kptpair.kpt1
        kpt2 = kptpair.kpt2
        Q_G = kptpair.Q_G  # Fourier components of kpoint pair
        nG = len(Q_G)

        n_nmG = np.zeros((len(n_n), len(m_m), nG), qpd.dtype)

        for j, n in enumerate(n_n):
            Q_G = kptpair.Q_G
            with self.context.timer('conj'):
                ut1cc_R = kpt1.ut_nR[n - kpt1.na].conj()
            with self.context.timer('paw'):
                C1_aGi = pawcorr.multiply(kpt1.P_ani, band=n - kpt1.na)
                n_nmG[j] = cpd(ut1cc_R, C1_aGi, kpt2, qpd, Q_G, block=block)

        return n_nmG

    @timer('get_optical_pair_density_head')
    def get_optical_pair_density_head(self, qpd, kptpair, n_n, m_m,
                                      block=False):
        """Get the optical limit of the pair density head (G=0) for a k-pair.
        """
        assert np.allclose(qpd.q_c, 0.0), f"{qpd.q_c} is not the optical limit"

        kpt1 = kptpair.kpt1
        kpt2 = kptpair.kpt2

        # v = (x, y, z)
        n_nmv = np.zeros((len(n_n), len(m_m), 3), qpd.dtype)

        for j, n in enumerate(n_n):
            n_nmv[j] = self.calculate_optical_pair_density_head(n, m_m,
                                                                kpt1, kpt2,
                                                                block=block)

        return n_nmv

    @timer('Calculate pair-densities')
    def calculate_pair_density(self, ut1cc_R, C1_aGi, kpt2, qpd, Q_G,
                               block=True):
        """Calculate FFT of pair-densities and add PAW corrections.

        ut1cc_R: 3-d complex ndarray
            Complex conjugate of the periodic part of the left hand side
            wave function.
        C1_aGi: list of ndarrays
            PAW corrections for all atoms.
        kpt2: KPoint object
            Right hand side k-point object.
        qpd: SingleQPWDescriptor
            Plane-wave descriptor for q=k2-k1.
        Q_G: 1-d int ndarray
            Mapping from flattened 3-d FFT grid to 0.5(G+q)^2<ecut sphere.
        """
        dv = qpd.gd.dv
        n_mG = qpd.empty(kpt2.blocksize)
        myblocksize = kpt2.nb - kpt2.na

        for ut_R, n_G in zip(kpt2.ut_nR, n_mG):
            n_R = ut1cc_R * ut_R
            with self.context.timer('fft'):
                n_G[:] = qpd.fft(n_R, 0, Q_G) * dv
        # PAW corrections:
        with self.context.timer('gemm'):
            for C1_Gi, P2_mi in zip(C1_aGi, kpt2.P_ani):
                # gemm(1.0, C1_Gi, P2_mi, 1.0, n_mG[:myblocksize], 't')
                mmm(1.0, P2_mi, 'N', C1_Gi, 'T', 1.0, n_mG[:myblocksize])

        if not block or self.blockcomm.size == 1:
            return n_mG
        else:
            n_MG = qpd.empty(kpt2.blocksize * self.blockcomm.size)
            with self.context.timer('all_gather'):
                self.blockcomm.all_gather(n_mG, n_MG)
            return n_MG[:kpt2.n2 - kpt2.n1]

    @timer('Optical limit')
    def calculate_optical_pair_velocity(self, n, kpt1, kpt2, block=False):
        # This has the effect of caching at most one kpoint.
        # This caching will be efficient only if we are looping over kpoints
        # in a particular way.
        #
        # It would be better to refactor so this caching is handled explicitly
        # by the caller providing the right thing.
        #
        # See https://gitlab.com/gpaw/gpaw/-/issues/625
        if self.ut_sKnvR is None or kpt1.K not in self.ut_sKnvR[kpt1.s]:
            self.ut_sKnvR = self.calculate_derivatives(kpt1)

        gd = self.gs.gd
        k_v = 2 * np.pi * np.dot(kpt1.k_c, np.linalg.inv(gd.cell_cv).T)

        ut_vR = self.ut_sKnvR[kpt1.s][kpt1.K][n - kpt1.n1]
        atomdata_a = self.gs.pawdatasets
        C_avi = [np.dot(atomdata.nabla_iiv.T, P_ni[n - kpt1.na])
                 for atomdata, P_ni in zip(atomdata_a, kpt1.P_ani)]

        blockbands = kpt2.nb - kpt2.na
        n0_mv = np.empty((kpt2.blocksize, 3), dtype=complex)
        nt_m = np.empty(kpt2.blocksize, dtype=complex)
        n0_mv[:blockbands] = -self.gs.gd.integrate(ut_vR,
                                                   kpt2.ut_nR).T
        nt_m[:blockbands] = self.gs.gd.integrate(kpt1.ut_nR[n - kpt1.na],
                                                 kpt2.ut_nR)

        n0_mv[:blockbands] += (1j * nt_m[:blockbands, np.newaxis] *
                               k_v[np.newaxis, :])

        for C_vi, P_mi in zip(C_avi, kpt2.P_ani):
            # gemm(1.0, C_vi, P_mi, 1.0, n0_mv[:blockbands], 'c')
            mmm(1.0, P_mi, 'N', C_vi, 'C', 1.0, n0_mv[:blockbands])

        if block and self.blockcomm.size > 1:
            n0_Mv = np.empty((kpt2.blocksize * self.blockcomm.size, 3),
                             dtype=complex)
            with self.context.timer('all_gather optical'):
                self.blockcomm.all_gather(n0_mv, n0_Mv)
            n0_mv = n0_Mv[:kpt2.n2 - kpt2.n1]

        return -1j * n0_mv

    def calculate_optical_pair_density_head(self, n, m_m, kpt1, kpt2,
                                            block=False):
        # Numerical threshold for the optical limit k dot p perturbation
        # theory expansion:
        threshold = 1

        eps1 = kpt1.eps_n[n - kpt1.n1]
        deps_m = (eps1 - kpt2.eps_n)[m_m - kpt2.n1]
        n0_mv = self.calculate_optical_pair_velocity(n, kpt1, kpt2,
                                                     block=block)

        deps_m = deps_m.copy()
        deps_m[deps_m == 0.0] = np.inf

        smallness_mv = np.abs(-1e-3 * n0_mv / deps_m[:, np.newaxis])
        inds_mv = (np.logical_and(np.inf > smallness_mv,
                                  smallness_mv > threshold))
        n0_mv *= - 1 / deps_m[:, np.newaxis]
        n0_mv[inds_mv] = 0

        return n0_mv

    @timer('Intraband')
    def intraband_pair_density(self, kpt, n_n):
        """Calculate intraband matrix elements of nabla"""
        # Bands and check for block parallelization
        na, nb, n1 = kpt.na, kpt.nb, kpt.n1
        vel_nv = np.zeros((nb - na, 3), dtype=complex)
        assert np.max(n_n) < nb, 'This is too many bands'
        assert np.min(n_n) >= na, 'This is too few bands'

        # Load kpoints
        gd = self.gs.gd
        k_v = 2 * np.pi * np.dot(kpt.k_c, np.linalg.inv(gd.cell_cv).T)
        atomdata_a = self.gs.pawdatasets

        # Break bands into degenerate chunks
        degchunks_cn = []  # indexing c as chunk number
        for n in n_n:
            inds_n = np.nonzero(np.abs(kpt.eps_n[n - n1] -
                                       kpt.eps_n) < 1e-5)[0] + n1

            # Has this chunk already been computed?
            oldchunk = any([n in chunk for chunk in degchunks_cn])
            if not oldchunk:
                if not all([ind in n_n for ind in inds_n]):
                    raise RuntimeError(
                        'You are cutting over a degenerate band '
                        'using block parallelization.')
                degchunks_cn.append((inds_n))

        # Calculate matrix elements by diagonalizing each block
        for ind_n in degchunks_cn:
            deg = len(ind_n)
            ut_nvR = self.gs.gd.zeros((deg, 3), complex)
            vel_nnv = np.zeros((deg, deg, 3), dtype=complex)
            # States are included starting from kpt.na
            ut_nR = kpt.ut_nR[ind_n - na]

            # Get derivatives
            for ind, ut_vR in zip(ind_n, ut_nvR):
                ut_vR[:] = self.make_derivative(kpt.s, kpt.K,
                                                ind, ind + 1)[0]

            # Treat the whole degenerate chunk
            for n in range(deg):
                ut_vR = ut_nvR[n]
                C_avi = [np.dot(atomdata.nabla_iiv.T, P_ni[ind_n[n] - na])
                         for atomdata, P_ni in zip(atomdata_a,
                                                   kpt.P_ani)]

                nabla0_nv = -self.gs.gd.integrate(ut_vR, ut_nR).T
                nt_n = self.gs.gd.integrate(ut_nR[n], ut_nR)
                nabla0_nv += 1j * nt_n[:, np.newaxis] * k_v[np.newaxis, :]

                for C_vi, P_ni in zip(C_avi, kpt.P_ani):
                    # gemm(1.0, C_vi, P_ni[ind_n - na], 1.0, nabla0_nv, 'c')
                    mmm(1.0, P_ni[ind_n - na], 'N', C_vi, 'C', 1.0, nabla0_nv)

                vel_nnv[n] = -1j * nabla0_nv

            for iv in range(3):
                vel, _ = np.linalg.eig(vel_nnv[..., iv])
                vel_nv[ind_n - na, iv] = vel  # Use eigenvalues

        return vel_nv[n_n - na]

    def calculate_derivatives(self, kpt):
        ut_sKnvR = [{}, {}]
        ut_nvR = self.make_derivative(kpt.s, kpt.K, kpt.n1, kpt.n2)
        ut_sKnvR[kpt.s][kpt.K] = ut_nvR

        return ut_sKnvR

    @timer('Derivatives')
    def make_derivative(self, s, K, n1, n2):
        gs = self.gs
        U_cc = gs.ibz2bz[K].U_cc
        A_cv = gs.gd.cell_cv
        M_vv = np.dot(np.dot(A_cv.T, U_cc.T), np.linalg.inv(A_cv).T)
        ik = gs.kd.bz2ibz_k[K]
        assert gs.kd.comm.size == 1
        kpt = gs.kpt_qs[ik][s]
        psit_nG = kpt.psit_nG
        iG_Gv = 1j * gs.pd.get_reciprocal_vectors(q=ik, add_q=False)
        ut_nvR = gs.gd.zeros((n2 - n1, 3), complex)
        for n in range(n1, n2):
            for v in range(3):
                ut_R = gs.ibz2bz[K].map_pseudo_wave(
                    gs.pd.ifft(iG_Gv[:, v] * psit_nG[n], ik))
                for v2 in range(3):
                    ut_nvR[n - n1, v2] += ut_R * M_vv[v, v2]

        return ut_nvR


def phase_shifted_fft_indices(k1_c, k2_c, qpd, coordinate_transformation=None):
    """Get phase shifted FFT indices for G-vectors inside the cutoff sphere.

    The output 1D FFT indices Q_G can be used to extract the plane-wave
    components G of the phase shifted Fourier transform

    n_kk'(G+q) = FFT_G[e^(-i[k+q-k']r) n_kk'(r)]

    where n_kk'(r) is some lattice periodic function and the wave vector
    difference k + q - k' is commensurate with the reciprocal lattice.
    """
    N_c = qpd.gd.N_c
    Q_G = qpd.Q_qG[0]
    q_c = qpd.q_c
    if coordinate_transformation:
        q_c = coordinate_transformation(q_c)

    shift_c = k1_c + q_c - k2_c
    assert np.allclose(shift_c.round(), shift_c)
    shift_c = shift_c.round().astype(int)

    if shift_c.any() or coordinate_transformation:
        # Get the 3D FFT grid indices (relative reciprocal space coordinates)
        # of the G-vectors inside the cutoff sphere
        i_cG = np.unravel_index(Q_G, N_c)
        if coordinate_transformation:
            i_cG = coordinate_transformation(i_cG)
        # Shift the 3D FFT grid indices to account for the Bloch-phase shift
        # e^(-i[k+q-k']r)
        i_cG += shift_c[:, np.newaxis]
        # Transform back the FFT grid to 1D FFT indices
        Q_G = np.ravel_multi_index(i_cG, N_c, 'wrap')

    return Q_G


def get_gs_and_context(calc, txt, world, timer):
    """Interface to initialize gs and context from old input arguments.
    Should be phased out in the future!"""
    from gpaw.calculator import GPAW as OldGPAW
    from gpaw.new.ase_interface import ASECalculator as NewGPAW

    context = ResponseContext(txt=txt, timer=timer, comm=world)

    if isinstance(calc, (OldGPAW, NewGPAW)):
        assert calc.wfs.world.size == 1
        gs = calc.gs_adapter()
    else:
        gs = ResponseGroundStateAdapter.from_gpw_file(calc, context=context)

    return gs, context



import numpy as np

import ase.units

from gpaw.lrtddft2.ks_singles import KohnShamSingleExcitation
from gpaw.lrtddft2.lr_layouts import LrTDDFPTSolveLayout


class LrResponse:
    def __init__(self,
                 lrtddft2,
                 excitation_energy,
                 field_vector,
                 lorentzian_width,
                 sl_lrtddft=None):
        self.lrtddft2 = lrtddft2
        self.omega = excitation_energy
        self.eta = lorentzian_width
        self.field_vector = np.array(field_vector)
        self.sl_lrtddft = sl_lrtddft

        self.C_re = None
        self.C_im = None

        self.response_wf_ready = False

    def read(self):
        raise RuntimeError('Error in LrtddftResponseWF read: Not implemented')

    # TD-DFPT:
    #
    # A C = -Vext
    #
    # ( E+KN+w    KN     -eta     0    ) ( C_+w,Re )   ( -dm_laser )
    # (   KN    E+KN-w    0      -eta  ) ( C_-w,Re ) = ( -dm_laser )
    # (  eta      0     E+KN+w   -KN   ) ( C_+w,Im )   (      0    )
    # (   0      eta     -KN    E+KN-w ) ( C_-w,Im )   (      0    )
    #
    # N is occupation difference matrix (just like in Casida)
    #    (at least K = 0 gives density diffs for both real and imaginary part)
    def solve(self):
        if self.sl_lrtddft is None:
            self.solve_serial()
        else:
            self.solve_parallel()

    def get_response_coefficients(self, units='eVang'):
        C_re = self.C_re * 1.
        C_im = self.C_im * 1.

        if units == 'au':
            pass
        elif units == 'eVang':
            # FIXME: where these units come from???
            # S = 2 omega detla n C_im dm
            # units: 1/eV = eV * e ang * C
            # => units of C: 1/(eV * eV * e ang)
            # maybe
            # psi(w) = psi0(w) + E(w).mu psi1(w) + ...
            # E(w).mu units = V/ang e ang = eV ?
            # nope...
            C_re *= 1. / (ase.units.Hartree**2 * ase.units.Bohr)
            C_im *= 1. / (ase.units.Hartree**2 * ase.units.Bohr)
        else:
            raise RuntimeError(
                'Error in get_response_coefficients: Invalid units.')

        return (C_re, C_im)

    def get_response_data(self, units='eVangcgs'):
        """Get response data.

        Returns matrix where transitions are in rows and columns are
        occupied index, unoccupied index, occupied KS eigenvalue, unoccupied
        KS eigenvalue, occupied occupation number, unoccupied occupation
        number,
        real part of response coefficent, imaginary part of response
        coefficient, dipole moment x, y, and z-components, magnetic
        moment x, y,
        and z-components.
        """

        data = np.zeros((len(self.lrtddft2.ks_singles.kss_list),
                         2 + 2 + 2 + 2 + 3 + 3 + 3))
        kpt_ind = self.lrtddft2.kpt_ind

        for (ip, kss_ip) in enumerate(self.lrtddft2.ks_singles.kss_list):
            i = kss_ip.occ_ind
            p = kss_ip.unocc_ind

            eps_i = self.lrtddft2.calc.wfs.kpt_u[kpt_ind].eps_n[i]
            eps_p = self.lrtddft2.calc.wfs.kpt_u[kpt_ind].eps_n[p]

            f_i = self.lrtddft2.calc.wfs.kpt_u[kpt_ind].f_n[i]
            f_p = self.lrtddft2.calc.wfs.kpt_u[kpt_ind].f_n[p]

            data[ip, 0] = i
            data[ip, 1] = p

            data[ip, 2] = eps_i
            data[ip, 3] = eps_p

            data[ip, 4] = f_i
            data[ip, 5] = f_p

            data[ip, 6] = self.C_re[ip]
            data[ip, 7] = self.C_im[ip]

            data[ip, 8] = kss_ip.dip_mom_r[0]
            data[ip, 9] = kss_ip.dip_mom_r[1]
            data[ip, 10] = kss_ip.dip_mom_r[2]

            data[ip, 11] = kss_ip.magn_mom[0]
            data[ip, 12] = kss_ip.magn_mom[1]
            data[ip, 13] = kss_ip.magn_mom[2]

            df = f_i - f_p

            C_im_ip = self.C_im[ip]
            data[ip, 14] = df * C_im_ip**2

            dm = np.vdot(kss_ip.dip_mom_r, self.field_vector)
            data[ip, 15] = 2. * self.omega * df * C_im_ip * dm
            # data[ip,15] = 2. * (eps_p-eps_i) * df * C_im_ip * dm

            # FIXME: IS THIS CORRECT ???
            # maybe answer is here
            # Varsano et al., Phys.Chem.Chem.Phys. 11, 4481 (2009)
            mm = np.vdot(kss_ip.magn_mom, self.field_vector)
            data[ip, 16] = -df * C_im_ip * mm

        if units == 'au':
            pass
        elif units == 'eVangcgs':
            data[:, 2] *= ase.units.Hartree
            data[:, 3] *= ase.units.Hartree
            data[:, 6] *= 1. / (ase.units.Hartree**2 * ase.units.Bohr)
            data[:, 7] *= 1. / (ase.units.Hartree**2 * ase.units.Bohr)
            data[:, 8] *= ase.units.Bohr
            data[:, 9] *= ase.units.Bohr
            data[:, 10] *= ase.units.Bohr
            data[:, 11] *= float('NaN')  # FIXME: units?
            data[:, 12] *= float('NaN')  # FIXME: units?
            data[:, 13] *= float('NaN')  # FIXME: units?
            data[:, 14] *= float('NaN')  # FIXME: units?
            data[:, 15] *= float('NaN')  # FIXME: units?
            data[:, 16] *= float('NaN')  # FIXME: units?
            raise RuntimeError(
                'Error in get_response_data: Unit conversion from '
                'atomic units to eV ang cgs not implemented yet.'
            )
        else:
            raise RuntimeError('Error in get_response_data: Invalid units.')

        return data

    # amplitude, dipole, rotatory
    # dipole should integrate to absorption spectrum
    # rotatory should integrate to CD spectrum
    def get_transition_contribution_maps(self,
                                         occ_min_energy,
                                         occ_max_energy,
                                         unocc_min_energy,
                                         unocc_max_energy,
                                         width=0.05,
                                         energy_step=0.05,
                                         units='eVangcgs'):
        data = self.get_response_data(units='au')

        eps_i = data[:, 2]
        eps_p = data[:, 3]
        a_ip = data[:, 14]
        s_ip = data[:, 15]
        r_ip = data[:, 16]

        # occupied energy range
        wx = np.arange(occ_min_energy, occ_max_energy, energy_step)
        # unoccupied energy range
        wy = np.arange(unocc_min_energy, unocc_max_energy, energy_step)

        A = np.outer(wx, wy) * 0.
        S = np.outer(wx, wy) * 0.
        R = np.outer(wx, wy) * 0.

        # A(omega_x, omega_y) =
        #   sum_ip delta n_ip |C^(im)_ip|**2
        #            * g(omega_x - eps_i) g(omega_y - eps_p)
        #
        # D(omega_x, omega_y) =
        #   sum_ip 2 * delta n_ip C^(im)_ip * mu_ip
        #            * g(omega_x - eps_i) g(omega_y - eps_p)
        #
        # where g(x) is gaussian
        for (ei, ep, a, s, r) in zip(eps_i, eps_p, a_ip, s_ip, r_ip):
            gx = np.exp((-.5 / width / width) *
                        np.power(wx - ei, 2)) / width / np.sqrt(2 * np.pi)
            gy = np.exp((-.5 / width / width) *
                        np.power(wy - ep, 2)) / width / np.sqrt(2 * np.pi)

            A += a * np.outer(gx, gy)
            S += s * np.outer(gx, gy)
            R += r * np.outer(gx, gy)

        # FIXME: units of A, S, and R

        if units == 'au':
            pass
        elif units == 'eVangcgs':
            raise RuntimeError(
                'Error in get_transition_contribution_maps: Unit conversion '
                'from atomic units to eV ang cgs not implemented yet.'
            )
        else:
            raise RuntimeError(
                'Error in get_transition_contribution_maps: Invalid units.')

        return (wx, wy, A, S, R)

    def get_induced_density(self,
                            units='au',
                            collect=False,
                            amplitude_filter=1e-5):
        # Init pair densities
        dnt_Gip = self.lrtddft2.calc.wfs.gd.empty()
        dnt_gip = self.lrtddft2.calc.density.finegd.empty()
        drhot_gip = self.lrtddft2.calc.density.finegd.empty()

        drhot_g = self.lrtddft2.calc.density.finegd.empty()
        drhot_g[:] = 0.0

        C_im = self.C_im

        maxC = np.max(abs(C_im))

        for (ip, kss_ip) in enumerate(self.lrtddft2.ks_singles.kss_list):
            # distribute ips over eh comm
            if self.lrtddft2.lr_comms.get_local_eh_index(ip) is None:
                continue
            if abs(C_im[ip]) < amplitude_filter * maxC:
                continue
            dnt_Gip[:] = 0.0
            dnt_gip[:] = 0.0
            drhot_gip[:] = 0.0
            kss_ip.calculate_pair_density(dnt_Gip, dnt_gip, drhot_gip)
            # ?FIXME?: SHOULD HERE BE OCCUPATION DIFFERENCE? yes ####
            drhot_g += kss_ip.pop_diff * C_im[ip] * drhot_gip

        # sum over eh comm
        self.lrtddft2.lr_comms.eh_comm.sum(drhot_g)

        if collect:
            drhot_g = self.lrtddft2.calc.density.finegd.collect(drhot_g)

        if units == 'au':
            pass
        elif units == 'ang':
            drhot_g *= 1. / ase.units.Bohr**3
        else:
            raise RuntimeError('Error in get_induced_density: Invalid units.')

        return drhot_g

    def get_approximate_electron_and_hole_densities(self,
                                                    units='au',
                                                    collect=False,
                                                    amplitude_filter=1e-4):
        # Init pair densities
        dnt_Gip = self.lrtddft2.calc.wfs.gd.empty()
        dnt_gip = self.lrtddft2.calc.density.finegd.empty()
        drhot_gip = self.lrtddft2.calc.density.finegd.empty()

        drhot_ge = self.lrtddft2.calc.density.finegd.empty()
        drhot_ge[:] = 0.0
        drhot_gh = self.lrtddft2.calc.density.finegd.empty()
        drhot_gh[:] = 0.0

        #
        # Sum of Slater determinants:
        # diag ip ip     => |c_ip|**2 ( |psi_p|**2 - |psi_i]**2 )
        # offdiag ip iq  =>   c_ip c_iq psi_p psi_q
        # offdiag ip jp  => - c_ip c_jp psi_i psi_j
        # offdiag ip jq  => 0
        #
        # remember both ip,jq  and jq,ip
        #

        # occupations for electron and hole
        f_n = self.lrtddft2.calc.wfs.kpt_u[self.lrtddft2.kpt_ind].f_n
        fe_n = np.zeros(len(f_n))
        fh_n = np.zeros(len(f_n))

        C_im = self.C_im

        maxC = np.max(abs(C_im))

        # diagonal ip,ip
        for (ip, kss_ip) in enumerate(self.lrtddft2.ks_singles.kss_list):
            if self.lrtddft2.lr_comms.get_local_eh_index(ip) is None:
                continue
            if abs(C_im[ip]) < amplitude_filter * maxC:
                continue

            # decrease in density
            fh_n[kss_ip.occ_ind] -= C_im[ip] * C_im[ip] * kss_ip.pop_diff
            # increase in density
            fe_n[kss_ip.unocc_ind] += C_im[ip] * C_im[ip] * kss_ip.pop_diff

        for k in range(len(f_n)):
            if (abs(fh_n[k]) < amplitude_filter * maxC * maxC
                    and abs(fe_n[k]) < amplitude_filter * maxC * maxC):
                continue

            dnt_Gip[:] = 0.0
            dnt_gip[:] = 0.0
            drhot_gip[:] = 0.0

            kss_ip = KohnShamSingleExcitation(self.lrtddft2.calc,
                                              self.lrtddft2.kpt_ind, k, k)
            kss_ip.calculate_pair_density(dnt_Gip, dnt_gip, drhot_gip)

            # fh_n < 0
            drhot_gh += fh_n[k] * drhot_gip
            # fe_n > 0
            drhot_ge += fe_n[k] * drhot_gip

        # offdiagonal
        for (ip, kss_ip) in enumerate(self.lrtddft2.ks_singles.kss_list):
            if self.lrtddft2.lr_comms.get_local_eh_index(ip) is None:
                continue
            if abs(C_im[ip]) < amplitude_filter * maxC:
                continue

            for (jq, kss_jq) in enumerate(self.lrtddft2.ks_singles.kss_list):
                # if diagonal, skip because already done
                if (kss_ip.occ_ind == kss_jq.occ_ind
                        and kss_ip.unocc_ind == kss_jq.unocc_ind):
                    continue

                # if both are different, then orthogonal
                if (kss_ip.occ_ind != kss_jq.occ_ind
                        and kss_ip.unocc_ind != kss_jq.unocc_ind):
                    continue

                if abs(C_im[ip] * C_im[jq]) < amplitude_filter * maxC * maxC:
                    continue

                # if occupied state is the same, then only electron density
                # changes (ie use kss_??.unocc_ind)
                # this gives direction, and should integrate to zero
                if (kss_ip.occ_ind == kss_jq.occ_ind
                        and kss_ip.unocc_ind != kss_jq.unocc_ind):
                    dnt_Gip[:] = 0.0
                    dnt_gip[:] = 0.0
                    drhot_gip[:] = 0.0

                    kss_pq = KohnShamSingleExcitation(self.lrtddft2.calc,
                                                      self.lrtddft2.kpt_ind,
                                                      kss_ip.unocc_ind,
                                                      kss_jq.unocc_ind)
                    kss_pq.calculate_pair_density(dnt_Gip, dnt_gip, drhot_gip)

                    drhot_ge += (C_im[ip] * C_im[jq] * np.sqrt(
                        kss_ip.pop_diff * kss_jq.pop_diff)) * drhot_gip

                # if occupied state is the same, then only hole density
                # changes (ie use kss_??.occ_ind)
                # this gives direction, and should integrate to zero
                if (kss_ip.occ_ind != kss_jq.occ_ind
                        and kss_ip.unocc_ind == kss_jq.unocc_ind):
                    dnt_Gip[:] = 0.0
                    dnt_gip[:] = 0.0
                    drhot_gip[:] = 0.0
                    kss_ij = KohnShamSingleExcitation(self.lrtddft2.calc,
                                                      self.lrtddft2.kpt_ind,
                                                      kss_ip.occ_ind,
                                                      kss_jq.occ_ind)
                    kss_ij.calculate_pair_density(dnt_Gip, dnt_gip, drhot_gip)

                    drhot_gh -= (C_im[ip] * C_im[jq] * np.sqrt(
                        kss_ip.pop_diff * kss_jq.pop_diff)) * drhot_gip

        self.lrtddft2.lr_comms.eh_comm.sum(drhot_ge)
        self.lrtddft2.lr_comms.eh_comm.sum(drhot_gh)

        drhot_geh = drhot_ge + drhot_gh

        # Ige  = self.lrtddft2.calc.density.finegd.integrate(drhot_ge)
        # Igh  = self.lrtddft2.calc.density.finegd.integrate(drhot_gh)
        # Igeh = self.lrtddft2.calc.density.finegd.integrate(drhot_geh)

        # if self.lrtddft2.lr_comms.parent_comm.rank == 0:
        #    print 'drho_ge ', Ige
        #    print 'drho_gh ', Igh
        #    print 'drho_geh', Igeh

        if collect:
            drhot_ge = self.lrtddft2.calc.density.finegd.collect(drhot_ge)
            drhot_gh = self.lrtddft2.calc.density.finegd.collect(drhot_gh)
            drhot_geh = self.lrtddft2.calc.density.finegd.collect(drhot_geh)

        if units == 'au':
            pass
        elif units == 'ang':
            drhot_geh *= 1. / ase.units.Bohr**3
            drhot_gh *= 1. / ase.units.Bohr**3
            drhot_ge *= 1. / ase.units.Bohr**3
        else:
            raise RuntimeError(
                'Error in get_approximate_electron_and_hole_densities: '
                'Invalid units.'
            )

        return (drhot_ge, drhot_gh, drhot_geh)

    def solve_serial(self):
        nrows = len(self.lrtddft2.ks_singles.kss_list)

        A_matrix = np.zeros((nrows * 4, nrows * 4))
        K_matrix = self.lrtddft2.K_matrix.values

        # add K N
        # slow but if you're using serial it's ok anyway
        for (ip, kss_ip) in enumerate(self.lrtddft2.ks_singles.kss_list):
            for (jq, kss_jq) in enumerate(self.lrtddft2.ks_singles.kss_list):
                lip = self.lrtddft2.lr_comms.get_local_eh_index(ip)
                ljq = self.lrtddft2.lr_comms.get_local_dd_index(jq)
                if lip is None or ljq is None:
                    continue

                A_matrix[ip * 4 + 0,
                         jq * 4 + 0] = K_matrix[lip, ljq] * kss_jq.pop_diff
                A_matrix[ip * 4 + 0,
                         jq * 4 + 1] = K_matrix[lip, ljq] * kss_jq.pop_diff
                A_matrix[ip * 4 + 1,
                         jq * 4 + 0] = K_matrix[lip, ljq] * kss_jq.pop_diff
                A_matrix[ip * 4 + 1,
                         jq * 4 + 1] = K_matrix[lip, ljq] * kss_jq.pop_diff

                A_matrix[ip * 4 + 2,
                         jq * 4 + 2] = K_matrix[lip, ljq] * kss_jq.pop_diff
                A_matrix[ip * 4 + 2,
                         jq * 4 + 3] = -K_matrix[lip, ljq] * kss_jq.pop_diff
                A_matrix[ip * 4 + 3,
                         jq * 4 + 2] = -K_matrix[lip, ljq] * kss_jq.pop_diff
                A_matrix[ip * 4 + 3,
                         jq * 4 + 3] = K_matrix[lip, ljq] * kss_jq.pop_diff

        # diagonal stuff: E, w, and eta
        for (ip, kss_ip) in enumerate(self.lrtddft2.ks_singles.kss_list):
            lip = self.lrtddft2.lr_comms.get_local_eh_index(ip)
            ljq = self.lrtddft2.lr_comms.get_local_dd_index(ip)
            if lip is None or ljq is None:
                continue

            # energy difference and excitation energy
            A_matrix[ip * 4 + 0, ip * 4 + 0] += kss_ip.energy_diff + self.omega
            A_matrix[ip * 4 + 1, ip * 4 + 1] += kss_ip.energy_diff - self.omega
            A_matrix[ip * 4 + 2, ip * 4 + 2] += kss_ip.energy_diff + self.omega
            A_matrix[ip * 4 + 3, ip * 4 + 3] += kss_ip.energy_diff - self.omega

            # life-time parameter
            A_matrix[ip * 4 + 0, ip * 4 + 2] += -self.eta
            A_matrix[ip * 4 + 1, ip * 4 + 3] += -self.eta
            A_matrix[ip * 4 + 2, ip * 4 + 0] += self.eta
            A_matrix[ip * 4 + 3, ip * 4 + 1] += self.eta

        # RHS
        V_rhs = np.zeros(nrows * 4)
        for (jq, kss_jq) in enumerate(self.lrtddft2.ks_singles.kss_list):
            ljq = self.lrtddft2.lr_comms.get_local_dd_index(jq)
            if ljq is None:
                continue

            # 2 cos(wt) = exp(+iwt) + exp(iwt)
            # fixme: no minus here???
            # -dm_laser = - (-er) cos(wt) ????
            V_rhs[jq * 4 + 0] = np.dot(kss_jq.dip_mom_r, self.field_vector)
            V_rhs[jq * 4 + 1] = np.dot(kss_jq.dip_mom_r, self.field_vector)
            V_rhs[jq * 4 + 2] = 0.
            V_rhs[jq * 4 + 3] = 0.

        # no transpose
        # A_matrix[:] = A_matrix.transpose()

        # solve A C = V
        self.lrtddft2.lr_comms.parent_comm.sum(A_matrix)
        self.lrtddft2.lr_comms.dd_comm.sum(V_rhs)

        # np.set_printoptions(precision=5, suppress=True)
        # print A_matrix
        C = np.linalg.solve(A_matrix, V_rhs)
        # print C

        # response wavefunction
        #   perturbation was exp(iwt) + exp(-iwt) = 2 cos(wt)
        #   => real part of the polarizability is propto 2 cos(wt)
        #   => imag part of the polarizability is propto 2 sin(wt)
        #
        # exp(iwt)
        # dn+ = phi-**h phi0 + phi0**h phi+ = phi0 (phi-**h + phi+)
        # exp(-iwt)
        # dn- = phi+**h phi0 + phi0**h phi- = phi0 (phi+**h + phi-) = dn+**h
        #
        # 2 cos(wt)
        # dn_2cos =    dn+ + dn-      =     dn+ + dn+**h  = Re[dn+] + Re[dn-]
        # 2 sin(wt)
        # dn_2sin = -i(dn+ - dn-)     = -i (dn+ - dn-**h) = Im[dn+] - Im[dn-]

        self.C_re = np.zeros(len(self.lrtddft2.ks_singles.kss_list))
        self.C_im = np.zeros(len(self.lrtddft2.ks_singles.kss_list))

        for (jq, kss_jq) in enumerate(self.lrtddft2.ks_singles.kss_list):
            # Re[phi-**h + phi_+]xo
            self.C_re[jq] = C[jq * 4 + 0] + C[jq * 4 + 1]
            # Im[phi-**h + phi_+]
            self.C_im[jq] = C[jq * 4 + 2] - C[jq * 4 + 3]

        # normalization... where it comes from? fourier (or lorentzian)...
        self.C_re *= 1. / np.pi
        self.C_im *= 1. / np.pi

        # anyway, you can check that
        #
        # osc_z(omega) / (pi eta) =
        #    2 * omega * sum_ip n_ip C^(im)_ip(omega) * mu^(z)_ip
        #
        # where osc_z, you can get from Casida
        # and 1/(pi eta) is because of Lorentzian folding

        # spectrum
        # S_z(omega) = 2 * omega sum_ip n_ip C^(im)_ip(omega) * mu^(z)_ip

        return (self.C_re, self.C_im)

    def solve_parallel(self):
        # total rows
        nrows = len(self.lrtddft2.ks_singles.kss_list)
        # local
        nlrows = self.lrtddft2.K_matrix.values.shape[0]
        nlcols = self.lrtddft2.K_matrix.values.shape[1]

        # create BLACS layout
        layout = LrTDDFPTSolveLayout(self.sl_lrtddft, nrows,
                                     self.lrtddft2.lr_comms)
        if (nrows <
                np.max([layout.mprocs, layout.nprocs]) * layout.block_size):
            raise RuntimeError(
                'Linear response matrix is not large enough for the given '
                'number of processes (or block size) in sl_lrtddft. Please, '
                'use less processes (or smaller block size).'
            )

        # build local part
        # NOTE: scalapack needs TRANSPOSED matrix!!!
        K_matrix = self.lrtddft2.K_matrix.values
        KN_matrix_T = np.zeros([nlcols, nlrows])
        A_matrix_T = np.zeros((nlcols * 4, nlrows * 4))

        KN_matrix_T[:] = K_matrix.T
        for (jq, kss_jq) in enumerate(self.lrtddft2.ks_singles.kss_list):
            ljq = self.lrtddft2.lr_comms.get_local_dd_index(jq)
            if ljq is None:
                continue
            KN_matrix_T[ljq, :] = K_matrix[:, ljq] * kss_jq.pop_diff

        lip_list = []
        ljq_list = []
        for (ip, kss_ip) in enumerate(self.lrtddft2.ks_singles.kss_list):
            lip = self.lrtddft2.lr_comms.get_local_eh_index(ip)
            ljq = self.lrtddft2.lr_comms.get_local_dd_index(ip)
            if lip is not None:
                lip_list.append(lip)
            if ljq is not None:
                ljq_list.append(ljq)

        lip4_list = np.array(lip_list) * 4

        # add K N
        for ljq in ljq_list:
            A_matrix_T[ljq * 4 + 0, lip4_list + 0] = KN_matrix_T[ljq, :]
            A_matrix_T[ljq * 4 + 0, lip4_list + 1] = KN_matrix_T[ljq, :]
            A_matrix_T[ljq * 4 + 1, lip4_list + 0] = KN_matrix_T[ljq, :]
            A_matrix_T[ljq * 4 + 1, lip4_list + 1] = KN_matrix_T[ljq, :]

            A_matrix_T[ljq * 4 + 2, lip4_list + 2] = KN_matrix_T[ljq, :]
            A_matrix_T[ljq * 4 + 2, lip4_list + 3] = -KN_matrix_T[ljq, :]
            A_matrix_T[ljq * 4 + 3, lip4_list + 2] = -KN_matrix_T[ljq, :]
            A_matrix_T[ljq * 4 + 3, lip4_list + 3] = KN_matrix_T[ljq, :]

        # diagonal stuff: E, w, and eta
        for (ip, kss_ip) in enumerate(self.lrtddft2.ks_singles.kss_list):
            lip = self.lrtddft2.lr_comms.get_local_eh_index(ip)
            ljq = self.lrtddft2.lr_comms.get_local_dd_index(ip)
            if lip is None or ljq is None:
                continue

            # energy difference and excitation energy
            A_matrix_T[ljq * 4 + 0,
                       lip * 4 + 0] += kss_ip.energy_diff + self.omega
            A_matrix_T[ljq * 4 + 1,
                       lip * 4 + 1] += kss_ip.energy_diff - self.omega
            A_matrix_T[ljq * 4 + 2,
                       lip * 4 + 2] += kss_ip.energy_diff + self.omega
            A_matrix_T[ljq * 4 + 3,
                       lip * 4 + 3] += kss_ip.energy_diff - self.omega

            # life-time parameter, transposed again!
            A_matrix_T[ljq * 4 + 0, lip * 4 + 2] += self.eta
            A_matrix_T[ljq * 4 + 1, lip * 4 + 3] += self.eta
            A_matrix_T[ljq * 4 + 2, lip * 4 + 0] += -self.eta
            A_matrix_T[ljq * 4 + 3, lip * 4 + 1] += -self.eta

        # RHS
        V_rhs = []
        for (jq, kss_jq) in enumerate(self.lrtddft2.ks_singles.kss_list):
            if (jq % self.lrtddft2.lr_comms.parent_comm.size !=
                self.lrtddft2.lr_comms.parent_comm.rank):
                continue

            # 2 cos(wt) = exp(+iwt) + exp(iwt)
            # fixme: no minus here???
            # -dm_laser = - (-er) cos(wt) ????
            V_rhs.append(np.dot(kss_jq.dip_mom_r, self.field_vector))
            V_rhs.append(np.dot(kss_jq.dip_mom_r, self.field_vector))
            V_rhs.append(0.)
            V_rhs.append(0.)

        V_rhs_T = np.array([V_rhs])

        # solve A C = V
        C = layout.solve(A_matrix_T, V_rhs_T)[0]

        # see solve serial for comments
        self.C_re = np.zeros(len(self.lrtddft2.ks_singles.kss_list))
        self.C_im = np.zeros(len(self.lrtddft2.ks_singles.kss_list))

        l = 0
        for (jq, kss_jq) in enumerate(self.lrtddft2.ks_singles.kss_list):
            if (jq % self.lrtddft2.lr_comms.parent_comm.size !=
                self.lrtddft2.lr_comms.parent_comm.rank):
                continue
            # Re[phi-**h + phi_+]
            self.C_re[jq] = C[l * 4 + 0] + C[l * 4 + 1]
            # Im[phi-**h + phi_+]
            self.C_im[jq] = C[l * 4 + 2] - C[l * 4 + 3]
            l += 1

        self.lrtddft2.lr_comms.parent_comm.sum(self.C_re)
        self.lrtddft2.lr_comms.parent_comm.sum(self.C_im)

        # normalization... where it comes from? fourier (or lorentzian)...
        self.C_re *= 1. / np.pi
        self.C_im *= 1. / np.pi

        return (self.C_re, self.C_im)



# Written by Lauri Lehtovaara, 2008

"""This module defines CSCG-class, which implements conjugate gradient
for complex symmetric matrices. Requires Numpy and GPAW's own BLAS."""

import numpy as np

from gpaw.utilities.blas import axpy
from gpaw.mpi import rank

from .base import BaseSolver


class CSCG(BaseSolver):
    """Conjugate gradient for complex symmetric matrices

    This class solves a set of linear equations A.x = b using conjugate
    gradient for complex symmetric matrices. The matrix A is a complex,
    symmetric, and non-singular matrix. The method requires only access
    to matrix-vector product A.x = b, which is called A.dot(x). Thus A
    must provide the member function dot(self,x,b), where x and b are
    complex arrays (numpy.array([], complex), and x is the known vector,
    and b is the result.

    Now x and b are multivectors, i.e., list of vectors.
    """

    def solve(self, A, x, b):
        if self.timer is not None:
            self.timer.start('CSCG')

        # number of vectors
        nvec = len(x)

        # r_0 = b - A x_0
        r = self.gd.zeros(nvec, dtype=complex)
        A.dot(-x, r)
        r += b

        p = self.gd.zeros(nvec, dtype=complex)
        q = self.gd.zeros(nvec, dtype=complex)
        z = self.gd.zeros(nvec, dtype=complex)

        alpha = np.zeros((nvec,), dtype=complex)
        beta = np.zeros((nvec,), dtype=complex)
        rho = np.zeros((nvec,), dtype=complex)
        rhop = np.zeros((nvec,), dtype=complex)
        scale = np.zeros((nvec,), dtype=complex)
        tmp = np.zeros((nvec,), dtype=complex)

        rhop[:] = 1.

        # Multivector dot product, a^T b, where ^T is transpose
        def multi_zdotu(s, x, y, nvec):
            for i in range(nvec):
                s[i] = x[i].ravel().dot(y[i].ravel())
                # s[i] = dotu(x[i],y[i])
            self.gd.comm.sum(s)
            return s

        # Multivector ZAXPY: a x + y => y
        def multi_zaxpy(a, x, y, nvec):
            for i in range(nvec):
                axpy(a[i] * (1 + 0J), x[i], y[i])

        # Multiscale: a x => x
        def multi_scale(a, x, nvec):
            for i in range(nvec):
                x[i] *= a[i]

        # scale = square of the norm of b
        multi_zdotu(scale, b, b, nvec)
        scale = np.abs(scale)

        # if scale < eps, then convergence check breaks down
        if (scale < self.eps).any():
            raise RuntimeError(
                "CSCG method detected underflow for squared norm of "
                "right-hand side (scale = %le < eps = %le)." %
                (scale, self.eps))

        # print 'Scale = ', scale

        slow_convergence_iters = 100

        for i in range(self.max_iter):
            # z_i = (M^-1.r)
            A.apply_preconditioner(r, z)

            # rho_i-1 = r^T z_i-1
            multi_zdotu(rho, r, z, nvec)

            # print 'Rho = ', rho

            # if i=1, p_i = r_i-1
            # else beta = (rho_i-1 / rho_i-2) (alpha_i-1 / omega_i-1)
            #      p_i = r_i-1 + b_i-1 (p_i-1 - omega_i-1 v_i-1)
            beta = rho / rhop

            # print 'Beta = ', beta

            # if abs(beta) / scale < eps, then CSCG breaks down
            if ((i > 0) and
                ((np.abs(beta) / scale) < self.eps).any()):
                raise RuntimeError(
                    "Conjugate gradient method failed "
                    "(abs(beta)=%le < eps = %le)." %
                    (np.min(np.abs(beta)), self.eps))

            # p = z + beta p
            multi_scale(beta, p, nvec)
            p += z

            # q = A.p
            A.dot(p, q)

            # alpha_i = rho_i-1 / (p^T q_i)
            multi_zdotu(alpha, p, q, nvec)
            alpha = rho / alpha

            # print 'Alpha = ', alpha

            # x_i = x_i-1 + alpha_i p_i
            multi_zaxpy(alpha, p, x, nvec)
            # r_i = r_i-1 - alpha_i q_i
            multi_zaxpy(-alpha, q, r, nvec)

            # if ( |r|^2 < tol^2 ) done
            multi_zdotu(tmp, r, r, nvec)
            if ((np.abs(tmp) / scale) < self.tol * self.tol).all():
                # print 'R2 of proc #', rank, '  = ' , tmp, \
                #     ' after ', i+1, ' iterations'
                break

            # print if slow convergence
            if ((i + 1) % slow_convergence_iters) == 0:
                print('R2 of proc #', rank, '  = ', tmp,
                      ' after ', i + 1, ' iterations')

            # finally update rho
            rhop[:] = rho

        # if max iters reached, raise error
        if (i >= self.max_iter - 1):
            raise RuntimeError(
                "Conjugate gradient method failed to converged "
                "within given number of iterations (= %d)." % self.max_iter)

        # done
        self.iterations = i + 1
        # print 'CSCG iterations = ', self.iterations

        if self.timer is not None:
            self.timer.stop('CSCG')

        return self.iterations
        # print self.iterations



import numpy as np
from ase.calculators.tip3p import TIP3P, rOH, angleHOH
from ase.optimize import FIRE
from ase.constraints import FixBondLengths
from gpaw.utilities.watermodel import FixBondLengthsWaterModel
from gpaw.utilities.watermodel import TIP3PWaterModel as TIP3Pc
from ase.data import s22

# check that all combinations of watermodel implementations
# and constraints give the same results


def test_rattle(in_tmp_dir):
    pairs = [(3 * i + j, 3 * i + (j + 1) % 3)
             for i in range(2)
             for j in [0, 1, 2]]

    ct = 0
    distances = np.zeros(4)
    for calc in [TIP3P(), TIP3Pc()]:
        for fix in [FixBondLengths, FixBondLengthsWaterModel]:
            atoms = s22.create_s22_system('Water_dimer')

            for m in [0, 3]:
                atoms.set_angle(m + 1, m, m + 2, angleHOH)
                atoms.set_distance(m, m + 1, rOH, fix=0)
                atoms.set_distance(m, m + 2, rOH, fix=0)
                atoms.set_angle(m + 1, m, m + 2, angleHOH)

            atoms.set_constraint(fix(pairs))

            atoms.calc = calc

            with FIRE(atoms, logfile='test.log') as opt:
                opt.run(0.05)
            p = atoms.get_positions()
            d = np.linalg.norm(p[0] - p[3])
            distances[ct] = d
            ct += 1

    assert (np.diff(distances) < 1e-6).all()



import pytest
import numpy as np

from ase.build import molecule

from gpaw import GPAW
from gpaw.lrtddft2 import LrTDDFT2
from gpaw.test import equal


@pytest.mark.lrtddft
def test_lrtddft2_H2O_lcao(in_tmp_dir):
    name = 'H2O-lcao'
    atoms = molecule('H2O')
    atoms.center(vacuum=4)

    # Ground state
    calc = GPAW(h=0.4, mode='lcao', basis='dzp', txt='%s-gs.out' % name,
                poissonsolver={'name': 'fd'},
                nbands=8, xc='LDA')
    atoms.calc = calc
    atoms.get_potential_energy()
    calc.write('%s.gpw' % name, mode='all')

    # LrTDDFT2
    calc = GPAW('%s.gpw' % name, txt='%s-lr.out' % name)
    lr = LrTDDFT2(name, calc, fxc='LDA')
    lr.calculate()
    results = lr.get_transitions()[0:2]

    if 0:
        np.set_printoptions(precision=10)
        refstr = repr(results)
        refstr = refstr.replace('array', 'np.array')
        # Format a pep-compatible output
        refstr = ' '.join(refstr.split())
        refstr = refstr.replace('[ ', '[')
        refstr = refstr.replace(', np', ',\n       np')
        refstr = refstr.replace(', ', ',\n                 ')
        print('ref = %s' % refstr)

    ref = (np.array([6.0832418565,
                     8.8741524165,
                     13.5935582927,
                     14.291759976,
                     15.9923574087,
                     16.9926770576,
                     17.6505087373,
                     17.6924715713,
                     24.092978682,
                     25.0027646421,
                     25.6208743025,
                     26.9649406073,
                     29.5294888977,
                     29.8439220428]),
           np.array([3.6910848669e-02,
                     5.4510953134e-24,
                     3.0995301001e-01,
                     2.1335662795e-02,
                     7.8725024176e-22,
                     6.3875174151e-02,
                     1.0288169027e-01,
                     1.7216407337e-01,
                     2.8901383040e-02,
                     3.9354412451e-01,
                     2.1917641289e-02,
                     6.7746017219e-01,
                     7.9561718249e-02,
                     1.0625830362e-02]))

    tol = 1e-3
    for r0, r1 in zip(results, ref):
        equal(r0, r1, tol)



# flake8: noqa
# Written by Lauri Lehtovaara, 2008
"""This module defines BiCGStab-class, which implements biconjugate
gradient stabilized method. Requires Numpy and GPAW's own BLAS."""

import numpy as np

from gpaw.utilities.blas import axpy
from gpaw.mpi import rank

from .base import BaseSolver


class BiCGStab(BaseSolver):
    """Biconjugate gradient stabilized method

    This class solves a set of linear equations A.x = b using biconjugate
    gradient stabilized method (BiCGStab). The matrix A is a general,
    non-singular matrix, e.g., it can be nonsymmetric, complex, and
    indefinite. The method requires only access to matrix-vector product
    A.x = b, which is called A.dot(x). Thus A must provide the member
    function dot(self,x,b), where x and b are complex arrays
    (numpy.array([], complex), and x is the known vector, and
    b is the result.

    Now x and b are multivectors, i.e., list of vectors.
    """

    def solve(self, A, x, b):
        if self.timer is not None:
            self.timer.start('BiCGStab')

        # number of vectors
        nvec = len(x)

        # r_0 = b - A x_0
        r = self.gd.zeros(nvec, dtype=complex)
        A.dot(-x, r)
        r += b

        q = self.gd.zeros(nvec, dtype=complex)
        q[:] = r

        p = self.gd.zeros(nvec, dtype=complex)
        v = self.gd.zeros(nvec, dtype=complex)
        t = self.gd.zeros(nvec, dtype=complex)
        m = self.gd.zeros(nvec, dtype=complex)
        alpha = np.zeros((nvec, ), dtype=complex)
        rho = np.zeros((nvec, ), dtype=complex)
        rhop = np.zeros((nvec, ), dtype=complex)
        omega = np.zeros((nvec, ), dtype=complex)
        scale = np.zeros((nvec, ), dtype=complex)
        tmp = np.zeros((nvec, ), dtype=complex)

        rhop[:] = 1.
        omega[:] = 1.

        # Multivector dot product, a^H b, where ^H is conjugate transpose
        def multi_zdotc(s, x, y, nvec):
            for i in range(nvec):
                s[i] = np.vdot(x[i], y[i])
            self.gd.comm.sum(s)
            return s

        # Multivector ZAXPY: a x + y => y
        def multi_zaxpy(a, x, y, nvec):
            for i in range(nvec):
                axpy(a[i] * (1 + 0J), x[i], y[i])

        # Multiscale: a x => x
        def multi_scale(a, x, nvec):
            for i in range(nvec):
                x[i] *= a[i]

        # scale = square of the norm of b
        multi_zdotc(scale, b, b, nvec)
        scale = np.abs(scale)

        # if scale < eps, then convergence check breaks down
        if (scale < self.eps).any():
            raise RuntimeError(
                "BigCGStab method detected underflow for squared norm of right-hand side (scale = %le < eps = %le)."
                % (scale, self.eps))

        #print 'Scale = ', scale

        slow_convergence_iters = 50

        for i in range(self.max_iter):
            # rho_i-1 = q^H r_i-1
            multi_zdotc(rho, q, r, nvec)

            #print 'Rho = ', rho

            # if i=1, p_i = r_i-1
            # else beta = (rho_i-1 / rho_i-2) (alpha_i-1 / omega_i-1)
            #      p_i = r_i-1 + b_i-1 (p_i-1 - omega_i-1 v_i-1)
            beta = (rho / rhop) * (alpha / omega)

            #print 'Beta = ', beta

            # if abs(beta) / scale < eps, then BiCGStab breaks down
            if ((i > 0) and ((np.abs(beta) / scale) < self.eps).any()):
                raise RuntimeError(
                    "Biconjugate gradient stabilized method failed (abs(beta)=%le < eps = %le)."
                    % (np.min(np.abs(beta)), self.eps))

            # p = r + beta * (p - omega * v)
            # vvvvvvvvvvvvvvvvvvvvvvvvvvvvvv
            multi_zaxpy(-omega, v, p, nvec)
            multi_scale(beta, p, nvec)
            p += r
            # ^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^

            # v_i = A.(M^-1.p)
            A.apply_preconditioner(p, m)
            A.dot(m, v)
            # alpha_i = rho_i-1 / (q^H v_i)
            multi_zdotc(alpha, q, v, nvec)
            alpha = rho / alpha
            # s = r_i-1 - alpha_i v_i
            multi_zaxpy(-alpha, v, r, nvec)
            # s is denoted by r

            #print 'Alpha = ', alpha

            # x_i = x_i-1 + alpha_i (M^-1.p_i) + omega_i (M^-1.s)
            # next line is x_i = x_i-1 + alpha (M^-1.p_i)
            multi_zaxpy(alpha, m, x, nvec)

            # if ( |s|^2 < tol^2 ) done
            multi_zdotc(tmp, r, r, nvec)
            if ((np.abs(tmp) / scale) < self.tol * self.tol).all():
                #print 'R2 of proc #', rank, '  = ' , tmp, \
                #    ' after ', i+1, ' iterations'
                break

            # print if slow convergence
            if ((i + 1) % slow_convergence_iters) == 0:
                print('Log10 S2 of proc #', rank, '  = ' , np.round(np.log10(np.abs(tmp)),1), \
                      ' after ', i+1, ' iterations')

            # t = A.(M^-1.s), M = 1
            A.apply_preconditioner(r, m)
            A.dot(m, t)
            # omega_i = t^H s / (t^H t)
            multi_zdotc(omega, t, r, nvec)
            multi_zdotc(tmp, t, t, nvec)
            omega = omega / tmp

            #print 'Omega = ', omega

            # x_i = x_i-1 + alpha_i (M^-1.p_i) + omega_i (M^-1.s)
            # next line is x_i = ... + omega_i (M^-1.s)
            multi_zaxpy(omega, m, x, nvec)
            # r_i = s - omega_i * t
            multi_zaxpy(-omega, t, r, nvec)
            # s is no longer denoted by r

            # if ( |r|^2 < tol^2 ) done
            multi_zdotc(tmp, r, r, nvec)
            if ((np.abs(tmp) / scale) < self.tol * self.tol).all():
                #print 'R2 of proc #', rank, '  = ' , tmp, \
                #    ' after ', i+1, ' iterations'
                break

            # print if slow convergence
            if ((i + 1) % slow_convergence_iters) == 0:
                print('Log10 R2 of proc #', rank, '  = ' , np.round(np.log10(np.abs(tmp)),1), \
                      ' after ', i+1, ' iterations')

            # if abs(omega) < eps, then BiCGStab breaks down
            if ((np.abs(omega) / scale) < self.eps).any():
                raise RuntimeError(
                    "Biconjugate gradient stabilized method failed (abs(omega)/scale=%le < eps = %le)."
                    % (np.min(np.abs(omega)) / scale, self.eps))
            # finally update rho
            rhop[:] = rho

        # if max iters reached, raise error
        if (i >= self.max_iter - 1):
            raise RuntimeError(
                "Biconjugate gradient stabilized method failed to converged within given number of iterations (= %d)."
                % self.max_iter)

        # done
        self.iterations = i + 1
        #print 'BiCGStab iterations = ', self.iterations

        if self.timer is not None:
            self.timer.stop('BiCGStab')

        return self.iterations
        #print self.iterations



"""
Atomic Density Functional Theory
"""

from math import log, pi, sqrt

import numpy as np
from ase.data import atomic_names
from ase.utils import IOContext

from gpaw import ConvergenceError
from gpaw.atom.configurations import configurations
from gpaw.atom.radialgd import AERadialGridDescriptor
from gpaw.utilities import hartree
from gpaw.xc import XC

# fine-structure constant
alpha = 1 / 137.036


class AllElectron(IOContext):
    """Object for doing an atomic DFT calculation."""

    def __init__(self, symbol, xcname='LDA', scalarrel=True,
                 corehole=None, configuration=None, nofiles=True,
                 txt='-', gpernode=150, orbital_free=False, tw_coeff=1.):
        """Do an atomic DFT calculation.

        Example::

          a = AllElectron('Fe')
          a.run()
        """

        self.txt = self.openfile(txt)

        self.symbol = symbol
        self.xcname = xcname
        self.scalarrel = scalarrel
        self.nofiles = nofiles

        # Get reference state:
        self.Z, nlfe_j = configurations[symbol]

        # Collect principal quantum numbers, angular momentum quantum
        # numbers, occupation numbers and eigenvalues (j is a combined
        # index for n and l):
        self.n_j = [n for n, l, f, e in nlfe_j]
        self.l_j = [l for n, l, f, e in nlfe_j]
        self.f_j = [f for n, l, f, e in nlfe_j]
        self.e_j = [e for n, l, f, e in nlfe_j]

        if configuration is not None:
            j = 0
            for conf in configuration.split(','):
                if conf[0].isdigit():
                    n = int(conf[0])
                    l = 'spdf'.find(conf[1])
                    if len(conf) == 2:
                        f = 1.0
                    else:
                        f = float(conf[2:])
                    assert n == self.n_j[j]
                    assert l == self.l_j[j]
                    self.f_j[j] = f
                    j += 1
                else:
                    j += {'He': 1,
                          'Ne': 3,
                          'Ar': 5,
                          'Kr': 8,
                          'Xe': 11}[conf]

        maxnodes = max([n - l - 1 for n, l in zip(self.n_j, self.l_j)])
        self.N = (maxnodes + 1) * gpernode
        self.beta = 0.4

        self.orbital_free = orbital_free
        self.tw_coeff = tw_coeff

        if self.orbital_free:
            self.n_j = [1]
            self.l_j = [0]
            self.f_j = [self.Z]
            self.e_j = [self.e_j[-1]]

        t = self.text
        t()
        if scalarrel:
            t('Scalar-relativistic atomic ', end='')
        else:
            t('Atomic ', end='')
        t('%s calculation for %s (%s, Z=%d)' % (xcname, symbol,
                                                atomic_names[self.Z], self.Z))

        if corehole is not None:
            self.ncorehole, self.lcorehole, self.fcorehole = corehole

            # Find j for core hole and adjust occupation:
            for j in range(len(self.f_j)):
                if (self.n_j[j] == self.ncorehole and
                    self.l_j[j] == self.lcorehole):
                    assert self.f_j[j] == 2 * (2 * self.lcorehole + 1)
                    self.f_j[j] -= self.fcorehole
                    self.jcorehole = j
                    break

            coreholestate = '%d%s' % (self.ncorehole, 'spdf'[self.lcorehole])
            t('Core hole in %s state (%s occupation: %.1f)' % (
                coreholestate, coreholestate, self.f_j[self.jcorehole]))
        else:
            self.jcorehole = None
            self.fcorehole = 0

    def text(self, *args, **kwargs):
        self.txt.write(kwargs.get('sep', ' ').join([str(arg)
                                                    for arg in args]) +
                       kwargs.get('end', '\n'))

    def initialize_wave_functions(self):
        r = self.r
        dr = self.dr
        # Initialize with Slater function:
        for l, e, u in zip(self.l_j, self.e_j, self.u_j):
            if self.symbol in ['Hf', 'Ta', 'W', 'Re', 'Os',
                               'Ir', 'Pt', 'Au']:
                a = sqrt(-4.0 * e)
            else:
                a = sqrt(-2.0 * e)

            u[:] = r**(1 + l) * np.exp(-a * r)
            norm = np.dot(u**2, dr)
            u *= 1.0 / sqrt(norm)

    def run(self):
        #     beta g
        # r = ------, g = 0, 1, ..., N - 1
        #     N - g
        #
        #        rN
        # g = --------
        #     beta + r

        t = self.text
        N = self.N
        beta = self.beta
        t(N, 'radial gridpoints.')
        self.rgd = AERadialGridDescriptor(beta / N, 1.0 / N, N)
        g = np.arange(N, dtype=float)
        self.r = self.rgd.r_g
        self.dr = self.rgd.dr_g
        self.d2gdr2 = self.rgd.d2gdr2()

        # Number of orbitals:
        nj = len(self.n_j)

        # Radial wave functions multiplied by radius:
        self.u_j = np.zeros((nj, self.N))

        # Effective potential multiplied by radius:
        self.vr = np.zeros(N)

        # Electron density:
        self.n = np.zeros(N)

        # Always spinpaired nspins=1
        self.xc = XC(self.xcname)

        # Initialize for non-local functionals
        if self.xc.type == 'GLLB':
            self.xc.initialize_1d(self)

        n_j = self.n_j
        l_j = self.l_j
        f_j = self.f_j
        e_j = self.e_j

        Z = self.Z    # nuclear charge
        r = self.r    # radial coordinate
        dr = self.dr  # dr/dg
        n = self.n    # electron density
        vr = self.vr  # effective potential multiplied by r

        vHr = np.zeros(self.N)
        self.vXC = np.zeros(self.N)

        self.initialize_wave_functions()
        n[:] = self.calculate_density()

        bar = '|------------------------------------------------|'
        t(bar)

        niter = 0
        nitermax = 117
        qOK = log(1e-10)
        mix = 0.4

        # orbital_free needs more iterations and coefficient
        if self.orbital_free:
            mix = 0.01
            nitermax = 2000
            e_j[0] /= self.tw_coeff
            if Z > 10:  # help convergence for third row elements
                mix = 0.002
                nitermax = 10000

        vrold = None

        while True:
            # calculate hartree potential
            hartree(0, n * r * dr, r, vHr)

            # add potential from nuclear point charge (v = -Z / r)
            vHr -= Z

            # calculated exchange correlation potential and energy
            self.vXC[:] = 0.0

            if self.xc.type == 'GLLB':
                # Update the potential to self.vXC an the energy to self.Exc
                Exc = self.xc.get_xc_potential_and_energy_1d(self.vXC)
            else:
                Exc = self.xc.calculate_spherical(self.rgd,
                                                  n.reshape((1, -1)),
                                                  self.vXC.reshape((1, -1)))

            # calculate new total Kohn-Sham effective potential and
            # admix with old version

            vr[:] = (vHr + self.vXC * r)

            if self.orbital_free:
                vr /= self.tw_coeff

            if niter > 0:
                vr[:] = mix * vr + (1 - mix) * vrold
            vrold = vr.copy()

            # solve Kohn-Sham equation and determine the density change
            self.solve()
            dn = self.calculate_density() - n
            n += dn

            # estimate error from the square of the density change integrated
            q = log(np.sum((r * dn)**2))

            # print progress bar
            if niter == 0:
                q0 = q
                b0 = 0
            else:
                b = int((q0 - q) / (q0 - qOK) * 50)
                if b > b0:
                    self.txt.write(bar[b0:min(b, 50)])
                    self.txt.flush()
                    b0 = b

            # check if converged and break loop if so
            if q < qOK:
                self.txt.write(bar[b0:])
                self.txt.flush()
                break

            niter += 1
            if niter > nitermax:
                raise RuntimeError('Did not converge!')

        tau = self.calculate_kinetic_energy_density()

        t()
        t('Converged in %d iteration%s.' % (niter, 's'[:niter != 1]))

        Ekin = 0

        for f, e in zip(f_j, e_j):
            Ekin += f * e

        e_coulomb = 2 * pi * np.dot(n * r * (vHr - Z), dr)
        Ekin += -4 * pi * np.dot(n * vr * r, dr)

        if self.orbital_free:
            # e and vr are not scaled back
            # instead Ekin is scaled for total energy
            # (printed and inside setup)
            Ekin *= self.tw_coeff
            t()
            t('Lambda:{0}'.format(self.tw_coeff))
            t('Correct eigenvalue:{0}'.format(e_j[0] * self.tw_coeff))
            t()

        t()
        t('Energy contributions:')
        t('-------------------------')
        t('Kinetic:   %+13.6f' % Ekin)
        t('XC:        %+13.6f' % Exc)
        t('Potential: %+13.6f' % e_coulomb)
        t('-------------------------')
        t('Total:     %+13.6f' % (Ekin + Exc + e_coulomb))
        self.ETotal = Ekin + Exc + e_coulomb
        t()

        t('state      eigenvalue         ekin         rmax')
        t('-----------------------------------------------')
        for m, l, f, e, u in zip(n_j, l_j, f_j, e_j, self.u_j):
            # Find kinetic energy:
            k = e - np.sum((np.where(abs(u) < 1e-160, 0, u)**2 *  # XXXNumeric!
                            vr * dr)[1:] / r[1:])

            # Find outermost maximum:
            g = self.N - 4
            while u[g - 1] >= u[g]:
                g -= 1
            x = r[g - 1:g + 2]
            y = u[g - 1:g + 2]
            A = np.transpose(np.array([x**i for i in range(3)]))
            c, b, a = np.linalg.solve(A, y)
            assert a < 0.0
            rmax = -0.5 * b / a

            s = 'spdf'[l]
            t('%d%s^%-4.1f: %12.6f %12.6f %12.3f' % (m, s, f, e, k, rmax))
        t('-----------------------------------------------')
        t('(units: Bohr and Hartree)')

        for m, l, u in zip(n_j, l_j, self.u_j):
            self.write(u, 'ae', n=m, l=l)

        self.write(n, 'n')
        self.write(vr, 'vr')
        self.write(vHr, 'vHr')
        self.write(self.vXC, 'vXC')
        self.write(tau, 'tau')

        self.Ekin = Ekin
        self.e_coulomb = e_coulomb
        self.Exc = Exc

    def write(self, array, name=None, n=None, l=None):
        if self.nofiles:
            return

        if name:
            name = self.symbol + '.' + name
        else:
            name = self.symbol

        if l is not None:
            assert n is not None
            if n > 0:
                # Bound state:
                name += '.%d%s' % (n, 'spdf'[l])
            else:
                name += '.x%d%s' % (-n, 'spdf'[l])

        f = open(name, 'w')
        for r, a in zip(self.r, array):
            print(r, a, file=f)

    def calculate_density(self):
        """Return the electron charge density divided by 4 pi"""
        n = np.dot(self.f_j,
                   np.where(abs(self.u_j) < 1e-160, 0,
                            self.u_j)**2) / (4 * pi)
        n[1:] /= self.r[1:]**2
        n[0] = n[1]
        return n

    def calculate_kinetic_energy_density(self):
        """Return the kinetic energy density"""
        return self.radial_kinetic_energy_density(self.f_j, self.l_j, self.u_j)

    def radial_kinetic_energy_density(self, f_j, l_j, u_j):
        """Kinetic energy density from a restricted set of wf's
        """
        shape = np.shape(u_j[0])
        dudr = np.zeros(shape)
        tau = np.zeros(shape)
        for f, l, u in zip(f_j, l_j, u_j):
            self.rgd.derivative(u, dudr)
            # contribution from angular derivatives
            if l > 0:
                tau += f * l * (l + 1) * np.where(abs(u) < 1e-160, 0, u)**2
            # contribution from radial derivatives
            dudr = u - self.r * dudr
            tau += f * np.where(abs(dudr) < 1e-160, 0, dudr)**2
        tau[1:] /= self.r[1:]**4
        tau[0] = tau[1]

        return 0.5 * tau / (4 * pi)

    def calculate_kinetic_energy_density2(self):
        """Return the kinetic energy density
        calculation over R(r)=u(r)/r
        slower convergence with # of radial grid points for
        Ekin of H than radial_kinetic_energy_density
        """

        shape = self.u_j.shape[1]
        R = np.zeros(shape)
        dRdr = np.zeros(shape)
        tau = np.zeros(shape)
        for f, l, u in zip(self.f_j, self.l_j, self.u_j):
            R[1:] = u[1:] / self.r[1:]
            if l == 0:
                # estimate value at origin by Taylor series to first order
                d1 = self.r[1]
                d2 = self.r[2]
                R[0] = .5 * (R[1] + R[2] + (R[1] - R[2]) *
                             (d1 + d2) / (d2 - d1))
            else:
                R[0] = 0
            self.rgd.derivative(R, dRdr)
            # contribution from radial derivatives
            tau += f * np.where(abs(dRdr) < 1e-160, 0, dRdr)**2
            # contribution from angular derivatives
            if l > 0:
                R[1:] = R[1:] / self.r[1:]
                if l == 1:
                    R[0] = R[1]
                else:
                    R[0] = 0
                tau += f * l * (l + 1) * np.where(abs(R) < 1e-160, 0, R)**2

        return 0.5 * tau / (4 * pi)

    def solve(self):
        """Solve the Schrodinger equation

        ::

             2
            d u     1  dv  du   u     l(l + 1)
          - --- - ---- -- (-- - -) + [-------- + 2M(v - e)] u = 0,
              2      2 dr  dr   r         2
            dr    2Mc                    r


        where the relativistic mass::

                   1
          M = 1 - --- (v - e)
                    2
                  2c

        and the fine-structure constant alpha = 1/c = 1/137.036
        is set to zero for non-scalar-relativistic calculations.

        On the logaritmic radial grids defined by::

              beta g
          r = ------,  g = 0, 1, ..., N - 1
              N - g

                 rN
          g = --------, r = [0; oo[
              beta + r

        the Schrodinger equation becomes::

           2
          d u      du
          --- c  + -- c  + u c  = 0
            2  2   dg  1      0
          dg

        with the vectors c0, c1, and c2  defined by::

                 2 dg 2
          c  = -r (--)
           2       dr

                  2         2
                 d g  2    r   dg dv
          c  = - --- r  - ---- -- --
           1       2         2 dr dr
                 dr       2Mc

                                    2    r   dv
          c  = l(l + 1) + 2M(v - e)r  + ---- --
           0                               2 dr
                                        2Mc
        """
        r = self.r
        dr = self.dr
        vr = self.vr

        c2 = -(r / dr)**2
        c10 = -self.d2gdr2 * r**2  # first part of c1 vector

        if self.scalarrel:
            self.r2dvdr = np.zeros(self.N)
            self.rgd.derivative(vr, self.r2dvdr)
            self.r2dvdr *= r
            self.r2dvdr -= vr
        else:
            self.r2dvdr = None

        # solve for each quantum state separately
        for j, (n, l, e, u) in enumerate(zip(self.n_j, self.l_j,
                                             self.e_j, self.u_j)):
            nodes = n - l - 1  # analytically expected number of nodes
            delta = -0.2 * e
            nn, A = shoot(u, l, vr, e, self.r2dvdr, r, dr, c10, c2,
                          self.scalarrel)
            # adjust eigenenergy until u has the correct number of nodes
            while nn != nodes:
                diff = np.sign(nn - nodes)
                while diff == np.sign(nn - nodes):
                    e -= diff * delta
                    nn, A = shoot(u, l, vr, e, self.r2dvdr, r, dr, c10, c2,
                                  self.scalarrel)
                delta /= 2

            # adjust eigenenergy until u is smooth at the turning point
            de = 1.0
            while abs(de) > 1e-9:
                norm = np.dot(np.where(abs(u) < 1e-160, 0, u)**2, dr)
                u *= 1.0 / sqrt(norm)
                de = 0.5 * A / norm
                x = abs(de / e)
                if x > 0.1:
                    de *= 0.1 / x
                e -= de
                assert e < 0.0
                nn, A = shoot(u, l, vr, e, self.r2dvdr, r, dr, c10, c2,
                              self.scalarrel)
            self.e_j[j] = e
            u *= 1.0 / sqrt(np.dot(np.where(abs(u) < 1e-160, 0, u)**2, dr))

    def solve_confined(self, j, rc, vconf=None):
        """Solve the Schroedinger equation in a confinement potential.

        Solves the Schroedinger equation like the solve method, but with a
        number of differences.  Before invoking this method, run solve() to
        get initial guesses.

        Parameters:
            j: solves only for the state given by j
            rc: solution cutoff. Solution will be zero outside this.
            vconf: added to the potential (use this as confinement potential)

        Returns: a tuple containing the solution u and its energy e.

        Unlike the solve method, this method will not alter any attributes of
        this object.
        """
        r = self.r
        dr = self.dr
        vr = self.vr.copy()
        if vconf is not None:
            vr += vconf * r

        c2 = -(r / dr)**2
        c10 = -self.d2gdr2 * r**2  # first part of c1 vector

        if j is None:
            n, l, e, u = 3, 2, -0.15, self.u_j[-1].copy()
        else:
            n = self.n_j[j]
            l = self.l_j[j]
            e = self.e_j[j]
            u = self.u_j[j].copy()

        nn, A = shoot_confined(u, l, vr, e, self.r2dvdr, r, dr, c10, c2,
                               self.scalarrel, rc=rc, beta=self.beta)
        assert nn == n - l - 1  # run() should have been called already

        # adjust eigenenergy until u is smooth at the turning point
        de = 1.0
        while abs(de) > 1e-9:
            norm = np.dot(np.where(abs(u) < 1e-160, 0, u)**2, dr)
            u *= 1.0 / sqrt(norm)
            de = 0.5 * A / norm
            x = abs(de / e)
            if x > 0.1:
                de *= 0.1 / x
            e -= de
            assert e < 0.0

            nn, A = shoot_confined(u, l, vr, e, self.r2dvdr, r, dr, c10, c2,
                                   self.scalarrel, rc=rc, beta=self.beta)
        u *= 1.0 / sqrt(np.dot(np.where(abs(u) < 1e-160, 0, u)**2, dr))
        return u, e

    def kin(self, l, u, e=None):  # XXX move to Generator
        r = self.r[1:]
        dr = self.dr[1:]

        c0 = 0.5 * l * (l + 1) / r**2
        c1 = -0.5 * self.d2gdr2[1:]
        c2 = -0.5 * dr**-2

        if e is not None and self.scalarrel:
            x = 0.5 * alpha**2
            Mr = r * (1.0 + x * e) - x * self.vr[1:]
            c0 += ((Mr - r) * (self.vr[1:] - e * r) +
                   0.5 * x * self.r2dvdr[1:] / Mr) / r**2
            c1 -= 0.5 * x * self.r2dvdr[1:] / (Mr * dr * r)

        fp = c2 + 0.5 * c1
        fm = c2 - 0.5 * c1
        f0 = c0 - 2 * c2
        kr = np.zeros(self.N)
        kr[1:] = f0 * u[1:] + fm * u[:-1]
        kr[1:-1] += fp[:-1] * u[2:]
        kr[0] = 0.0
        return kr

    def r2g(self, r):
        """Convert radius to index of the radial grid."""
        return int(r * self.N / (self.beta + r))

    def get_confinement_potential(self, alpha, ri, rc):
        r"""Create a smooth confinement potential.

        Returns a (potential) function which is zero inside the radius ri
        and goes to infinity smoothly at rc, after which point it is nan.
        The potential is given by::

                   alpha         /   rc - ri \
          V(r) = --------   exp ( - --------- )   for   ri < r < rc
                  rc - r         \    r - ri /

        """
        i_ri = self.r2g(ri)
        i_rc = self.r2g(rc)
        if self.r[i_rc] == rc:
            # Avoid division by zero in the odd case that rc coincides
            # exactly with a grid point (which actually happens sometimes)
            i_rc -= 1

        potential = np.zeros(np.shape(self.r))
        r = self.r[i_ri + 1:i_rc + 1]
        exponent = - (rc - ri) / (r - ri)
        denom = rc - r
        value = np.exp(exponent) / denom
        potential[i_ri + 1:i_rc + 1] = value
        potential[i_rc + 1:] = np.inf

        return alpha * potential

    def __del__(self):
        self.close()


def shoot(u, l, vr, e, r2dvdr, r, dr, c10, c2, scalarrel, gmax=None):
    """n, A = shoot(u, l, vr, e, ...)

    For guessed trial eigenenergy e, integrate the radial Schrodinger
    equation::

          2
         d u      du
         --- c  + -- c  + u c  = 0
           2  2   dg  1      0
         dg

               2 dg 2
        c  = -r (--)
         2       dr

                2         2
               d g  2    r   dg dv
        c  = - --- r  - ---- -- --
         1       2         2 dr dr
               dr       2Mc

                                  2    r   dv
        c  = l(l + 1) + 2M(v - e)r  + ---- --
         0                               2 dr
                                      2Mc

    The resulting wavefunction is returned in input vector u.
    The number of nodes of u is returned in attribute n.
    Returned attribute A, is a measure of the size of the derivative
    discontinuity at the classical turning point.
    The trial energy e is correct if A is zero and n is the correct number
    of nodes."""

    if scalarrel:
        x = 0.5 * alpha**2  # x = 1 / (2c^2)
        Mr = r * (1.0 + x * e) - x * vr
    else:
        Mr = r
    c0 = l * (l + 1) + 2 * Mr * (vr - e * r)
    if gmax is None and (c0 > 0).all():
        raise ConvergenceError('Bad initial electron density guess!')

    c1 = c10
    if scalarrel:
        c0 += x * r2dvdr / Mr
        c1 = c10 - x * r * r2dvdr / (Mr * dr)

    # vectors needed for numeric integration of diff. equation
    fm = 0.5 * c1 - c2
    fp = 0.5 * c1 + c2
    f0 = c0 - 2 * c2

    if gmax is None:
        # set boundary conditions at r -> oo (u(oo) = 0 is implicit)
        u[-1] = 1.0

        # perform backwards integration from infinity to the turning point
        g = len(u) - 2
        u[-2] = u[-1] * f0[-1] / fm[-1]
        while c0[g] > 0.0:  # this defines the classical turning point
            u[g - 1] = (f0[g] * u[g] + fp[g] * u[g + 1]) / fm[g]
            if u[g - 1] < 0.0:
                # There should't be a node here!  Use a more negative
                # eigenvalue:
                print('!!!!!!', end=' ')
                return 100, None
            if u[g - 1] > 1e100:
                u *= 1e-100
            g -= 1

        # stored values of the wavefunction and the first derivative
        # at the turning point
        gtp = g + 1
        utp = u[gtp]
        if gtp == len(u) - 1:
            return 100, 0.0
        dudrplus = 0.5 * (u[gtp + 1] - u[gtp - 1]) / dr[gtp]
    else:
        gtp = gmax

    # set boundary conditions at r -> 0
    u[0] = 0.0
    u[1] = 1.0

    # perform forward integration from zero to the turning point
    g = 1
    nodes = 0
    # integrate one step further than gtp
    # (such that dudr is defined in gtp)
    while g <= gtp:
        u[g + 1] = (fm[g] * u[g - 1] - f0[g] * u[g]) / fp[g]
        if u[g + 1] * u[g] < 0:
            nodes += 1
        g += 1
    if gmax is not None:
        return

    # scale first part of wavefunction, such that it is continuous at gtp
    u[:gtp + 2] *= utp / u[gtp]

    # determine size of the derivative discontinuity at gtp
    dudrminus = 0.5 * (u[gtp + 1] - u[gtp - 1]) / dr[gtp]
    A = (dudrplus - dudrminus) * utp

    return nodes, A


def shoot_confined(u, l, vr, e, r2dvdr, r, dr, c10, c2, scalarrel,
                   gmax=None, rc=10., beta=7.):
    """This method is used by the solve_confined method."""
    # XXX much of this is pasted from the ordinary shoot method

    if scalarrel:
        x = 0.5 * alpha**2  # x = 1 / (2c^2)
        Mr = r * (1.0 + x * e) - x * vr
    else:
        Mr = r
    c0 = l * (l + 1) + 2 * Mr * (vr - e * r)
    if gmax is None and (c0 > 0).all():
        print("""
Problem with initial electron density guess!  Try to run the program
with the '-n' option (non-scalar-relativistic calculation) and then
try again without the '-n' option (this will generate a good initial
guess for the density).
""")
        raise SystemExit
    c1 = c10
    if scalarrel:
        c0 += x * r2dvdr / Mr
        c1 = c10 - x * r * r2dvdr / (Mr * dr)

    # vectors needed for numeric integration of diff. equation
    fm = 0.5 * c1 - c2
    fp = 0.5 * c1 + c2
    f0 = c0 - 2 * c2

    if gmax is None:
        gcut = int(rc * len(r) / (beta + rc))
        # set boundary conditions at r -> oo (u(oo) = 0 is implicit)
        u[gcut - 1] = 1.
        u[gcut:] = 0.

        # perform backwards integration from infinity to the turning point
        g = gcut - 2
        u[g] = u[g + 1] * f0[g + 1] / fm[g + 1]

        while c0[g] > 0.0:  # this defines the classical turning point
            u[g - 1] = (f0[g] * u[g] + fp[g] * u[g + 1]) / fm[g]
            if u[g - 1] < 0.0:
                # There should't be a node here!  Use a more negative
                # eigenvalue:
                print('!!!!!!', end=' ')
                return 100, None
            if u[g - 1] > 1e100:
                u *= 1e-100
            g -= 1

        # stored values of the wavefunction and the first derivative
        # at the turning point
        gtp = g + 1
        utp = u[gtp]
        dudrplus = 0.5 * (u[gtp + 1] - u[gtp - 1]) / dr[gtp]
    else:
        gtp = gmax

    # set boundary conditions at r -> 0
    u[0] = 0.0
    u[1] = 1.0

    # perform forward integration from zero to the turning point
    g = 1
    nodes = 0
    # integrate one step further than gtp
    # (such that dudr is defined in gtp)
    while g <= gtp:
        u[g + 1] = (fm[g] * u[g - 1] - f0[g] * u[g]) / fp[g]
        if u[g + 1] * u[g] < 0:
            nodes += 1
        g += 1
    if gmax is not None:
        return

    # scale first part of wavefunction, such that it is continuous at gtp
    u[:gtp + 2] *= utp / u[gtp]

    # determine size of the derivative discontinuity at gtp
    dudrminus = 0.5 * (u[gtp + 1] - u[gtp - 1]) / dr[gtp]
    A = (dudrplus - dudrminus) * utp

    return nodes, A


if __name__ == '__main__':
    a = AllElectron('Cu', scalarrel=True)
    a.run()



import pytest
from gpaw.atom.aeatom import AllElectronAtom, c
from gpaw.test import equal


@pytest.mark.ci
def test_aeatom():
    Z = 79  # gold atom
    kwargs = dict(ngpts=5000, alpha2=1000 * Z**2, ngauss=200)

    # Test Schroedinger equation:
    aea = AllElectronAtom(Z, log=None)
    aea.initialize(**kwargs)

    errors = []
    for channel in aea.channels:
        channel.solve(-Z)
        for n in range(7):
            e = channel.e_n[n]
            e0 = -0.5 * Z**2 / (n + channel.l + 1)**2
            errors.append(abs(e / e0 - 1))
    print(max(errors))
    equal(max(errors), 0, 2.0e-5)

    # Test Dirac equation:
    aea = AllElectronAtom(Z, dirac=True, log=None)
    aea.initialize(**kwargs)

    errors = []
    for channel in aea.channels:
        channel.solve(-Z)
        for n in range(7):
            e = channel.e_n[n]
            if channel.k > 0:
                n += 1
            e0 = (1 +
                  (Z / c)**2 /
                  ((channel.k**2 - (Z / c)**2)**0.5 + n)**2)**-0.5 - 1
            e0 *= c**2
            errors.append(abs(e / e0 - 1))
    print(max(errors))
    equal(max(errors), 0, 4.0e-5)



"""Point Group Symmetry.

The code was originally written for the paper:

    S. Kaappa, S. Malola, H. Hakkinen

    J. Phys. Chem. A; vol. 122, 43, pp. 8576-8584 (2018)

"""

from .group import PointGroup
from .check import SymmetryChecker

__all__ = ['PointGroup', 'SymmetryChecker', 'point_group_names']

point_group_names = ['C2', 'C2v', 'C3v', 'D2d', 'D3h', 'D5', 'D5h',
                     'Ico', 'Ih', 'Oh', 'Td', 'Th']



import numpy as np
from math import pi
from gpaw.coulomb import Coulomb
from gpaw.grid_descriptor import GridDescriptor
from gpaw.mpi import world
from gpaw.utilities.gauss import coordinates
from gpaw.test import equal
import time


def test_coulomb():
    def test_coulomb(N=2**6, a=20):
        Nc = (N, N, N)            # Number of grid point
        gd = GridDescriptor(Nc, (a, a, a), True)  # grid-descriptor object
        # matrix with the square of the radial coordinate
        xyz, r2 = coordinates(gd)
        # matrix with the values of the radial coordinate
        r = np.sqrt(r2)
        nH = np.exp(-2 * r) / pi  # density of the hydrogen atom
        C = Coulomb(gd)           # coulomb calculator

        if world.size > 1:
            C.load('real')
            t0 = time.time()
            print('Processor %s of %s: %s Ha in %s sec' % (
                gd.comm.rank + 1,
                gd.comm.size,
                -0.5 * C.coulomb(nH, method='real'),
                time.time() - t0))
            return
        else:
            C.load('recip_ewald')
            C.load('recip_gauss')
            C.load('real')
            test = {}
            t0 = time.time()
            test['dual density'] = (-0.5 * C.coulomb(nH, nH.copy()),
                                    time.time() - t0)
            for method in ('real', 'recip_gauss', 'recip_ewald'):
                t0 = time.time()
                test[method] = (-0.5 * C.coulomb(nH, method=method),
                                time.time() - t0)
            return test

    analytic = -5 / 16.0
    res = test_coulomb(N=48, a=15)
    if world.size == 1:
        print('Units: Bohr and Hartree')
        print('%12s %8s %8s' % ('Method', 'Energy', 'Time'))
        print('%12s %2.6f %6s' % ('analytic', analytic, '--'))
        for method, et in res.items():
            print('%12s %2.6f %1.7f' % ((method,) + et))

        equal(res['real'][0], analytic, 6e-3)
        equal(res['recip_gauss'][0], analytic, 6e-3)
        equal(res['recip_ewald'][0], analytic, 2e-2)
        equal(res['dual density'][0], res['recip_gauss'][0], 1e-9)

    # mpirun -np 2 python coulomb.py --gpaw-parallel --gpaw-debug



import pytest
import numpy as np

from gpaw import GPAW
from gpaw.test import gen
from gpaw.xas import XAS, RecursionMethod
import gpaw.mpi as mpi


@pytest.mark.later
def test_corehole_si(in_tmp_dir, add_cwd_to_setup_paths, gpw_files):
    # restart from file
    calc = GPAW(gpw_files['si_corehole_pw'])
    si = calc.atoms

    if mpi.size == 1:
        xas = XAS(calc)
        x, y = xas.get_spectra()
    else:
        x = np.linspace(0, 10, 50)

    k = 2
    calc = calc.new(kpts=(k, k, k))
    calc.initialize(si)
    calc.set_positions(si)
    assert calc.wfs.dtype == complex

    r = RecursionMethod(calc)
    r.run(40)
    if mpi.size == 1:
        z = r.get_spectra(x)

    if 0:
        import pylab as p
        p.plot(x, y[0])
        p.plot(x, sum(y))
        p.plot(x, z[0])
        p.show()

    # 2p corehole
    s = gen('Si', name='hch2p', corehole=(2, 1, 0.5), gpernode=30)
    calc = GPAW(gpw_files['si_corehole_pw'],
                setups={0: s})
    si.calc = calc

    def stopcalc():
        calc.scf.converged = True

    calc.attach(stopcalc, 1)
    _ = si.get_potential_energy()


@pytest.mark.later
def test_si_nonortho(in_tmp_dir, add_cwd_to_setup_paths, gpw_files):
    # Generate setup for oxygen with half a core-hole:
    # restart from file
    # code moved to fixtures: si_corehole_sym,
    # si_corehole_nosym_pw, si_corehole_sym_pw
    calc1 = GPAW(gpw_files['si_corehole_sym_pw'])
    calc2 = GPAW(gpw_files['si_corehole_nosym_pw'])
    if mpi.size == 1:
        xas1 = XAS(calc1)
        x, y1 = xas1.get_spectra()
        xas2 = XAS(calc2)
        x2, y2 = xas2.get_spectra(E_in=x)

        assert (np.sum(abs(y1 - y2)[0, :500]**2) < 5e-9)
        assert (np.sum(abs(y1 - y2)[1, :500]**2) < 5e-9)
        assert (np.sum(abs(y1 - y2)[2, :500]**2) < 5e-9)



from __future__ import annotations

from types import ModuleType
from typing import TYPE_CHECKING, Generic, TypeVar, Callable

import gpaw.fftw as fftw
import numpy as np
from ase.io.ulm import NDArrayReader
from gpaw.core.domain import Domain
from gpaw.core.matrix import Matrix
from gpaw.mpi import MPIComm
from gpaw.typing import Array1D, Literal, Self, ArrayND

if TYPE_CHECKING:
    from gpaw.core.uniform_grid import UGArray, UGDesc

from gpaw.new import prod

DomainType = TypeVar('DomainType', bound=Domain)


class DistributedArrays(Generic[DomainType]):
    desc: DomainType

    def __init__(self,
                 dims: int | tuple[int, ...],
                 myshape: tuple[int, ...],
                 comm: MPIComm,
                 domain_comm: MPIComm,
                 data: np.ndarray | None,
                 dv: float,
                 dtype,
                 xp=None):
        self.myshape = myshape
        self.comm = comm
        self.domain_comm = domain_comm
        self.dv = dv

        # convert int to tuple:
        self.dims = dims if isinstance(dims, tuple) else (dims,)

        if self.dims:
            mydims0 = (self.dims[0] + comm.size - 1) // comm.size
            d1 = min(comm.rank * mydims0, self.dims[0])
            d2 = min((comm.rank + 1) * mydims0, self.dims[0])
            mydims0 = d2 - d1
            self.mydims = (mydims0,) + self.dims[1:]
        else:
            self.mydims = ()

        fullshape = self.mydims + self.myshape

        if data is not None:
            if data.shape != fullshape:
                raise ValueError(
                    f'Bad shape for data: {data.shape} != {fullshape}')
            if data.dtype != dtype:
                raise ValueError(
                    f'Bad dtype for data: {data.dtype} != {dtype}')
            if xp is not None:
                assert (xp is np) == isinstance(
                    data, (np.ndarray, NDArrayReader)), xp
        else:
            data = (xp or np).empty(fullshape, dtype)

        self.data = data
        self.xp: ModuleType
        if isinstance(data, (np.ndarray, NDArrayReader)):
            self.xp = np
        else:
            from gpaw.gpu import cupy as cp
            self.xp = cp
        self._matrix: Matrix | None = None

    def new(self, data=None) -> DistributedArrays:
        raise NotImplementedError

    def copy(self):
        return self.new(data=self.data.copy())

    def __getitem__(self, index):
        raise NotImplementedError

    def __iter__(self):
        for index in range(self.dims[0]):
            yield self[index]

    def flat(self):
        if self.dims == ():
            yield self
        else:
            for index in np.indices(self.dims).reshape((len(self.dims), -1)).T:
                yield self[tuple(index)]

    def to_xp(self, xp):
        if xp is self.xp:
            assert xp is np, 'cp -> cp should not be needed!'
            return self
        if xp is np:
            return self.new(data=self.xp.asnumpy(self.data))
        else:
            return self.new(data=xp.asarray(self.data))

    @property
    def matrix(self) -> Matrix:
        if self._matrix is not None:
            return self._matrix

        nx = prod(self.myshape)
        shape = (self.dims[0], prod(self.dims[1:]) * nx)
        myshape = (self.mydims[0], prod(self.mydims[1:]) * nx)
        dist = (self.comm, -1, 1)

        data = self.data.reshape(myshape)
        self._matrix = Matrix(*shape, data=data, dist=dist)

        return self._matrix

    def matrix_elements(self,
                        other: Self,
                        *,
                        out: Matrix | None = None,
                        symmetric: bool | Literal['_default'] = '_default',
                        function=None,
                        domain_sum=True,
                        cc: bool = False) -> Matrix:
        if symmetric == '_default':
            symmetric = self is other

        comm = self.comm

        if out is None:
            out = Matrix(self.dims[0], other.dims[0],
                         dist=(comm, -1, 1),
                         dtype=self.desc.dtype,
                         xp=self.xp)

        if comm.size == 1:
            assert other.comm.size == 1
            if function:
                assert symmetric
                other = function(other)

            M1 = self.matrix
            M2 = other.matrix
            out = M1.multiply(M2, opb='C', alpha=self.dv,
                              symmetric=symmetric, out=out)

            # Plane-wave expansion of real-valued functions needs a correction:
            self._matrix_elements_correction(M1, M2, out, symmetric)

        else:
            if symmetric:
                _parallel_me_sym(self, out, function)
            else:
                _parallel_me(self, other, out)

        if not cc:
            out.complex_conjugate()

        if domain_sum:
            self.domain_comm.sum(out.data)

        return out

    def _matrix_elements_correction(self,
                                    M1: Matrix,
                                    M2: Matrix,
                                    out: Matrix,
                                    symmetric: bool) -> None:
        """Hook for PlaneWaveExpansion."""
        pass

    def abs_square(self,
                   weights: Array1D,
                   out: UGArray) -> None:
        """Add weighted absolute square of data to output array.

        See also :xkcd:`849`.
        """
        raise NotImplementedError

    def add_ked(self,
                weights: Array1D,
                out: UGArray) -> None:
        """Add weighted absolute square of gradient of data to output array."""
        raise NotImplementedError

    def gather(self, out=None, broadcast=False):
        raise NotImplementedError

    def gathergather(self):
        a_xX = self.gather()  # gather X
        if a_xX is not None:
            m_xX = a_xX.matrix.gather()  # gather x
            if m_xX.dist.comm.rank == 0:
                data = m_xX.data
                if a_xX.data.dtype != data.dtype:
                    data = data.view(complex)
                return self.desc.new(comm=None).from_data(data)

    def scatter_from(self, data: ArrayND | None = None) -> None:
        raise NotImplementedError

    def redist(self,
               domain,
               comm1: MPIComm, comm2: MPIComm) -> DistributedArrays:
        result = domain.empty(self.dims)
        if comm1.rank == 0:
            a = self.gather()
        else:
            a = None
        if comm2.rank == 0:
            result.scatter_from(a)
        comm2.broadcast(result.data, 0)
        return result

    def interpolate(self,
                    plan1: fftw.FFTPlans | None = None,
                    plan2: fftw.FFTPlans | None = None,
                    grid: UGDesc | None = None,
                    out: UGArray | None = None) -> UGArray:
        raise NotImplementedError


def _parallel_me(psit1_nX: DistributedArrays,
                 psit2_nX: DistributedArrays,
                 M_nn: Matrix) -> None:

    comm = psit1_nX.comm
    nbands = psit1_nX.dims[0]

    psit1_nX = psit1_nX[:]

    B = (nbands + comm.size - 1) // comm.size

    n_r = [min(r * B, nbands) for r in range(comm.size + 1)]

    buf1_nX = psit1_nX.desc.empty(B)
    buf2_nX = psit1_nX.desc.empty(B)
    psit_nX = psit2_nX

    for shift in range(comm.size):
        rrequest = None
        srequest = None

        if shift < comm.size - 1:
            srank = (comm.rank + shift + 1) % comm.size
            rrank = (comm.rank - shift - 1) % comm.size
            n1 = n_r[rrank]
            n2 = n_r[rrank + 1]
            mynb = n2 - n1
            if mynb > 0:
                rrequest = comm.receive(buf1_nX.data[:mynb], rrank, 11, False)
            if psit2_nX.data.size > 0:
                srequest = comm.send(psit2_nX.data, srank, 11, False)

        r2 = (comm.rank - shift) % comm.size
        n1 = n_r[r2]
        n2 = n_r[r2 + 1]
        m_nn = psit1_nX.matrix_elements(psit_nX[:n2 - n1],
                                        cc=True, domain_sum=False)

        M_nn.data[:, n1:n2] = m_nn.data

        if rrequest:
            comm.wait(rrequest)
        if srequest:
            comm.wait(srequest)

        psit_nX = buf1_nX
        buf1_nX, buf2_nX = buf2_nX, buf1_nX


def _parallel_me_sym(psit1_nX: DistributedArrays,
                     M_nn: Matrix,
                     operator: None | Callable[[DistributedArrays],
                                               DistributedArrays]
                     ) -> None:
    """..."""
    comm = psit1_nX.comm
    nbands = psit1_nX.dims[0]
    B = (nbands + comm.size - 1) // comm.size
    mynbands = psit1_nX.mydims[0]

    n_r = [min(r * B, nbands) for r in range(comm.size + 1)]
    mynbands_r = [n_r[r + 1] - n_r[r] for r in range(comm.size)]
    assert mynbands_r[comm.rank] == mynbands

    psit2_nX = psit1_nX
    buf1_nX = psit1_nX.desc.empty(B)
    buf2_nX = psit1_nX.desc.empty(B)
    half = comm.size // 2

    for shift in range(half + 1):
        rrequest = None
        srequest = None

        if shift < half:
            srank = (comm.rank + shift + 1) % comm.size
            rrank = (comm.rank - shift - 1) % comm.size
            skip = comm.size % 2 == 0 and shift == half - 1
            rmynb = mynbands_r[rrank]
            if not (skip and comm.rank < half) and rmynb > 0:
                rrequest = comm.receive(buf1_nX.data[:rmynb], rrank, 11, False)
            if not (skip and comm.rank >= half) and psit1_nX.data.size > 0:
                srequest = comm.send(psit1_nX.data, srank, 11, False)

        if shift == 0:
            if operator is not None:
                op_psit1_nX = operator(psit1_nX)
            else:
                op_psit1_nX = psit1_nX
            op_psit1_nX = op_psit1_nX[:]  # local view

        if not (comm.size % 2 == 0 and shift == half and comm.rank < half):
            r2 = (comm.rank - shift) % comm.size
            n1 = n_r[r2]
            n2 = n_r[r2 + 1]
            m_nn = op_psit1_nX.matrix_elements(psit2_nX[:n2 - n1],
                                               symmetric=(shift == 0),
                                               cc=True, domain_sum=False)
            M_nn.data[:, n1:n2] = m_nn.data

        if rrequest:
            comm.wait(rrequest)
        if srequest:
            comm.wait(srequest)

        psit2_nX = buf1_nX
        buf1_nX, buf2_nX = buf2_nX, buf1_nX

    requests = []
    blocks = []
    nrows = (comm.size - 1) // 2
    for row in range(nrows):
        for column in range(comm.size - nrows + row, comm.size):
            if comm.rank == row:
                n1 = n_r[column]
                n2 = n_r[column + 1]
                if mynbands > 0 and n2 > n1:
                    requests.append(
                        comm.send(M_nn.data[:, n1:n2].T.conj().copy(),
                                  column, 12, False))
            elif comm.rank == column:
                n1 = n_r[row]
                n2 = n_r[row + 1]
                if mynbands > 0 and n2 > n1:
                    block = np.empty((mynbands, n2 - n1), M_nn.dtype)
                    blocks.append((n1, n2, block))
                    requests.append(comm.receive(block, row, 12, False))

    comm.waitall(requests)
    for n1, n2, block in blocks:
        M_nn.data[:, n1:n2] = block



import numpy as np
import pytest

from gpaw.gpu import cupy as cp
from gpaw.utilities.blas import (gpu_axpy, gpu_dotc, gpu_dotu, gpu_gemm,
                                 gpu_gemv, gpu_mmm, gpu_r2k, gpu_rk, gpu_scal,
                                 mmm, r2k, rk)


@pytest.mark.gpu
@pytest.mark.parametrize('dtype', [float, complex])
def test_blas(gpu, dtype):
    N = 100
    rng = np.random.default_rng(seed=42)
    a = np.zeros((N, N), dtype=dtype)
    b = np.zeros_like(a)
    c = np.zeros_like(a)
    x = np.zeros((N,), dtype=dtype)
    y = np.zeros_like(x)
    if dtype == float:
        a[:] = rng.random((N, N))
        b[:] = rng.random((N, N))
        c[:] = rng.random((N, N))
        x[:] = rng.random((N,))
        y[:] = rng.random((N,))
    else:
        a.real = rng.random((N, N))
        a.imag = rng.random((N, N))
        b.real = rng.random((N, N))
        b.imag = rng.random((N, N))
        c.real = rng.random((N, N))
        c.imag = rng.random((N, N))
        x.real = rng.random((N,))
        x.imag = rng.random((N,))
        y.real = rng.random((N,))
        y.imag = rng.random((N,))

    a_gpu = cp.asarray(a)
    b_gpu = cp.asarray(b)
    c_gpu = cp.asarray(c)
    x_gpu = cp.asarray(x)
    y_gpu = cp.asarray(y)

    # axpy
    y += 0.5 * x
    check_cpu = y.sum()

    gpu_axpy(0.5, x_gpu, y_gpu)
    check_gpu = y_gpu.sum().get()

    assert check_cpu == pytest.approx(check_gpu, rel=1e-14)

    # mmm
    mmm(0.5, a, 'N', b, 'N', 0.2, c)
    check_cpu = c.sum()

    gpu_mmm(0.5, a_gpu, 'N', b_gpu, 'N', 0.2, c_gpu)
    check_gpu = c_gpu.sum().get()

    assert check_cpu == pytest.approx(check_gpu, rel=1e-14)

    # gemm
    c *= 0.2
    c += 0.5 * b @ a
    check_cpu = c.sum()

    gpu_gemm(0.5, a_gpu, b_gpu, 0.2, c_gpu)
    check_gpu = c_gpu.sum().get()

    assert check_cpu == pytest.approx(check_gpu, rel=1e-14)

    # gemv
    y *= 0.2
    y += 0.5 * a @ x
    check_cpu = y.sum()

    gpu_gemv(0.5, a_gpu, x_gpu, 0.2, y_gpu)
    check_gpu = y_gpu.sum().get()

    assert check_cpu == pytest.approx(check_gpu, rel=1e-14)

    # rk
    rk(0.5, a, 0.2, c)
    check_cpu = c.sum()

    gpu_rk(0.5, a_gpu, 0.2, c_gpu)
    check_gpu = c_gpu.sum().get()

    assert check_cpu == pytest.approx(check_gpu, rel=1e-14)

    # r2k
    r2k(0.5, a, b, 0.2, c)
    check_cpu = c.sum()

    gpu_r2k(0.5, a_gpu, b_gpu, 0.2, c_gpu)
    check_gpu = c_gpu.sum().get()

    assert check_cpu == pytest.approx(check_gpu, rel=1e-14)

    # dotc
    check_cpu = x.conj() @ y

    check_gpu = gpu_dotc(x_gpu, y_gpu)

    assert check_cpu == pytest.approx(check_gpu, rel=1e-14)

    # dotu
    check_cpu = x @ y

    check_gpu = gpu_dotu(x_gpu, y_gpu)

    assert check_cpu == pytest.approx(check_gpu, rel=1e-14)

    # scal
    a *= 0.5
    check_cpu = a.sum()

    gpu_scal(0.5, a_gpu)
    check_gpu = a_gpu.sum().get()

    assert check_cpu == pytest.approx(check_gpu, rel=1e-14)



import numpy as np

from gpaw.atom.radialgd import fsbt, EquidistantRadialGridDescriptor as RGD


def test_maths_fsbt():
    N = 1024
    L = 50.0
    h = L / N
    alpha = 5.0
    r = np.linspace(0, L, N, endpoint=False)
    G = np.linspace(0, np.pi / h, N // 2 + 1)
    n = np.exp(-alpha * r**2)

    for l in range(7):
        f = fsbt(l, n * r**l, r, G)
        f0 = (np.pi**0.5 / alpha**(l + 1.5) / 2**l * G**l / 4 *
              np.exp(-G**2 / (4 * alpha)))
        tol = 3e-6 * 10**(-7 + l)
        print(l, abs(f - f0).max(), 'tol=', tol)
        assert abs(f - f0).max() < tol

    rgd = RGD(r[1], len(r))
    g, f = rgd.fft(n * r)
    f0 = 4 * np.pi**1.5 / alpha**1.5 / 4 * np.exp(-g**2 / 4 / alpha)
    assert abs(f - f0).max() < 1e-6

    # This is how to do the inverse FFT:
    ggd = RGD(g[1], len(g))
    r, f = ggd.fft(f * g)
    assert abs(np.exp(-alpha * r**2) - f / 8 / np.pi**3).max() < 2e-3



import warnings

import numpy as np
from scipy.special import erf

from gpaw.poisson import BasePoissonSolver, PoissonSolver, FDPoissonSolver
from gpaw.fd_operators import Laplace, Gradient
from gpaw.wfd_operators import WeightedFDOperator
from gpaw.utilities.gauss import Gaussian
from gpaw import PoissonConvergenceError
from gpaw.utilities.timing import NullTimer


class SolvationPoissonSolver(FDPoissonSolver):
    """Base class for Poisson solvers with spatially varying dielectric.

    The Poisson equation
    div(epsilon(r) grad phi(r)) = -4 pi rho(r)
    is solved.
    """

    def __init__(self, nn=3, relax='J', eps=2e-10, maxiter=1000,
                 remove_moment=None, use_charge_center=False):
        if remove_moment is not None:
            raise NotImplementedError(
                'Removing arbitrary multipole moments '
                'is not implemented for SolvationPoissonSolver!')
        if nn == 'M':
            raise NotImplementedError(
                'Mehrstellen stencil is not implemented '
                'for SolvationPoissonSolver!')
        FDPoissonSolver.__init__(self, nn, relax, eps, maxiter, remove_moment,
                                 use_charge_center=use_charge_center)

    def set_dielectric(self, dielectric):
        """Set the dielectric.

        Arguments:
        dielectric -- A Dielectric instance.
        """
        self.dielectric = dielectric

    def load_gauss(self, center=None):
        """Load compensating charge distribution for charged systems.

        See Appendix B of
        A. Held and M. Walter, J. Chem. Phys. 141, 174108 (2014).
        """
        # XXX Check if update is needed (dielectric changed)?
        epsr, dx_epsr, dy_epsr, dz_epsr = self.dielectric.eps_gradeps
        gauss = Gaussian(self.gd, center=center)
        rho_g = gauss.get_gauss(0)
        phi_g = gauss.get_gauss_pot(0)
        x, y, z = gauss.xyz
        fac = 2. * np.sqrt(gauss.a) * np.exp(-gauss.a * gauss.r2)
        fac /= np.sqrt(np.pi) * gauss.r2
        fac -= erf(np.sqrt(gauss.a) * gauss.r) / (gauss.r2 * gauss.r)
        fac *= 2.0 * 1.7724538509055159
        dx_phi_g = fac * x
        dy_phi_g = fac * y
        dz_phi_g = fac * z
        sp = dx_phi_g * dx_epsr + dy_phi_g * dy_epsr + dz_phi_g * dz_epsr
        rho = epsr * rho_g - 1. / (4. * np.pi) * sp
        invnorm = np.sqrt(4. * np.pi) / self.gd.integrate(rho)
        self.phi_gauss = phi_g * invnorm
        self.rho_gauss = rho * invnorm


class WeightedFDPoissonSolver(SolvationPoissonSolver):
    """Weighted finite difference Poisson solver with dielectric.

    Following V. M. Sanchez, M. Sued and D. A. Scherlis,
    J. Chem. Phys. 131, 174108 (2009).
    """

    def create_laplace(self, gd, scale=1.0, n=1, dtype=float):
        operators = [Laplace(gd, scale, n, dtype)]
        operators += [Gradient(gd, j, scale, n, dtype) for j in (0, 1, 2)]
        return WeightedFDOperator(operators)

    def solve(self, phi, rho, charge=None,
              maxcharge=1e-6,
              zero_initial_phi=False, timer=None):
        self._init()
        if self.gd.pbc_c.all():
            actual_charge = self.gd.integrate(rho)
            if abs(actual_charge) > maxcharge:
                raise NotImplementedError(
                    'charged periodic systems are not implemented')
        self.restrict_op_weights()
        ret = FDPoissonSolver.solve(self, phi, rho, charge, maxcharge,
                                    zero_initial_phi)
        return ret

    def restrict_op_weights(self):
        """Restric operator weights to coarse grids."""
        self._init()
        weights = [self.dielectric.eps_gradeps] + self.op_coarse_weights
        for i, res in enumerate(self.restrictors):
            for j in range(4):
                res.apply(weights[i][j], weights[i + 1][j])
        self.step = 0.66666666 / self.operators[0].get_diagonal_element()

    def get_description(self):
        description = SolvationPoissonSolver.get_description(self)
        return description.replace(
            'solver with',
            'weighted FD solver with dielectric and')

    def _init(self):
        if self._initialized:
            return
        self.operators[0].set_weights(self.dielectric.eps_gradeps)
        self.op_coarse_weights = []
        for operator in self.operators[1:]:
            weights = [gd.empty() for gd in (operator.gd, ) * 4]
            self.op_coarse_weights.append(weights)
            operator.set_weights(weights)
        return SolvationPoissonSolver._init(self)


class PolarizationPoissonSolver(BasePoissonSolver):
    """Poisson solver with dielectric.

    Calculates the polarization charges first using only the
    vacuum poisson equation, then solves the vacuum equation
    with polarization charges.
    """

    def __init__(self, nn=3, relax='J', eps=2e-10, maxiter=1000,
                 remove_moment=None, use_charge_center=False,
                 gas_phase_poisson='fast'):
        self.nn = nn
        self.eps = eps
        self.maxiter = maxiter
        self.phi_tilde = None

        self.gas_phase_poisson = PoissonSolver(
            name=gas_phase_poisson, nn=nn, eps=eps,
            remove_moment=remove_moment,
            use_charge_center=use_charge_center)

    def set_dielectric(self, dielectric):
        """Set the dielectric.

        Arguments:
        dielectric -- A Dielectric instance.
        """
        self.dielectric = dielectric

    def get_description(self):
        return ('PolarizationPoissonSolver based on '
                + self.gas_phase_poisson.get_description())

    def set_grid_descriptor(self, gd):
        self.gd = gd
        self.gas_phase_poisson.set_grid_descriptor(gd)

    def solve(self, phi, rho, charge=None,
              maxcharge=1e-6,
              zero_initial_phi=False, timer=NullTimer()):
        # get initial meaningful phi -> only do this if necessary
        if zero_initial_phi:
            niter = self.gas_phase_poisson.solve(
                phi, rho, charge=None, maxcharge=maxcharge,
                zero_initial_phi=zero_initial_phi, timer=timer)
        else:
            niter = 0

        phi_old = phi.copy()
        while niter < self.maxiter:
            rho_mod = self.rho_with_polarization_charge(phi, rho)
            niter += self.gas_phase_poisson.solve(
                phi, rho_mod, charge=None, maxcharge=maxcharge,
                zero_initial_phi=zero_initial_phi, timer=timer)
            residual = phi - phi_old
            error = self.gd.comm.sum(np.dot(residual.ravel(),
                                            residual.ravel())) * self.gd.dv
            if error < self.eps:
                return niter
            phi_old = phi.copy()
            
        raise PoissonConvergenceError(
            'PolarizationPoisson solver did not converge in '
            + f'{niter} iterations!')

    def rho_with_polarization_charge(self, phi, rho):
        epsr, dx_epsr, dy_epsr, dz_epsr = self.dielectric.eps_gradeps
        dx_phi = self.gd.empty()
        dy_phi = self.gd.empty()
        dz_phi = self.gd.empty()
        Gradient(self.gd, 0, 1.0, self.nn).apply(phi, dx_phi)
        Gradient(self.gd, 1, 1.0, self.nn).apply(phi, dy_phi)
        Gradient(self.gd, 2, 1.0, self.nn).apply(phi, dz_phi)

        scalar_product = (
            dx_epsr * dx_phi +
            dy_epsr * dy_phi +
            dz_epsr * dz_phi)

        return (rho + scalar_product / (4. * np.pi)) / epsr

    def estimate_memory(self, mem):
        # XXX estimate your own contribution
        return self.gas_phase_poisson.estimate_memory(mem)

    def todict(self):
        my_dict = self.gas_phase_poisson.todict()
        my_dict['name'] = f"polarization-{my_dict['name']}"
        return my_dict


class ADM12PoissonSolver(SolvationPoissonSolver):
    """Poisson solver with dielectric.

    Following O. Andreussi, I. Dabo, and N. Marzari,
    J. Chem. Phys. 136, 064102 (2012).

    Warning: Not intended for production use, as it is not tested
             thouroughly!

    XXX TODO : * Correction for charged systems???
               * Check: Can the polarization charge introduce a monopole?
               * Convergence problems depending on eta. Apparently this
                 method works best with FFT as in the original Paper.
               * Optimize numerics.
    """

    def __init__(self, nn=3, relax='J', eps=2e-10, maxiter=1000,
                 remove_moment=None, eta=.6, use_charge_center=False):
        """Constructor for ADM12PoissonSolver.

        Additional arguments not present in SolvationPoissonSolver:
        eta -- linear mixing parameter
        """
        adm12_warning = UserWarning(
            'ADM12PoissonSolver is not tested thoroughly'
            ' and therefore not recommended for production code!')
        warnings.warn(adm12_warning)
        self.eta = eta
        SolvationPoissonSolver.__init__(
            self, nn, relax, eps, maxiter, remove_moment,
            use_charge_center=use_charge_center)

    def set_grid_descriptor(self, gd):
        SolvationPoissonSolver.set_grid_descriptor(self, gd)
        self.gradx = Gradient(gd, 0, 1.0, self.nn)
        self.grady = Gradient(gd, 1, 1.0, self.nn)
        self.gradz = Gradient(gd, 2, 1.0, self.nn)

    def get_description(self):
        if len(self.operators) == 0:
            return 'uninitialized ADM12PoissonSolver'
        else:
            description = SolvationPoissonSolver.get_description(self)
            return description.replace(
                'solver with',
                'ADM12 solver with dielectric and')

    def _init(self):
        if self._initialized:
            return
        self.rho_iter = self.gd.zeros()
        self.d_phi = self.gd.empty()
        return SolvationPoissonSolver._init(self)

    def solve(self, phi, rho, charge=None,
              maxcharge=1e-6,
              zero_initial_phi=False, timer=None):
        self._init()
        if self.gd.pbc_c.all():
            actual_charge = self.gd.integrate(rho)
            if abs(actual_charge) > maxcharge:
                raise NotImplementedError(
                    'charged periodic systems are not implemented')
        return FDPoissonSolver.solve(
            self, phi, rho, charge, maxcharge, zero_initial_phi,
            timer=timer)

    def solve_neutral(self, phi, rho, timer=None):
        self._init()
        self.rho = rho
        return SolvationPoissonSolver.solve_neutral(self, phi, rho)

    def iterate2(self, step, level=0):
        self._init()
        if level == 0:
            epsr, dx_epsr, dy_epsr, dz_epsr = self.dielectric.eps_gradeps
            self.gradx.apply(self.phis[0], self.d_phi)
            sp = dx_epsr * self.d_phi
            self.grady.apply(self.phis[0], self.d_phi)
            sp += dy_epsr * self.d_phi
            self.gradz.apply(self.phis[0], self.d_phi)
            sp += dz_epsr * self.d_phi
            self.rho_iter = self.eta / (4. * np.pi) * sp + \
                (1. - self.eta) * self.rho_iter
            self.rhos[0][:] = (self.rho_iter + self.rho) / epsr
        return SolvationPoissonSolver.iterate2(self, step, level)



from ase import Atoms
from gpaw import GPAW, PoissonSolver
from gpaw.test import equal
from ase.units import Bohr, Hartree


def test_xc_revPBE():
    a = 7.5 * Bohr
    n = 16
    atoms = Atoms('He', [(0.0, 0.0, 0.0)], cell=(a, a, a), pbc=True)
    params = dict(mode='fd',
                  gpts=(n, n, n),
                  nbands=1,
                  xc={'name': 'PBE', 'stencil': 1},
                  poissonsolver=PoissonSolver('fd'))
    atoms.calc = GPAW(**params)
    e1 = atoms.get_potential_energy()
    e1a = atoms.calc.get_reference_energy()
    params['xc'] = {'name': 'revPBE', 'stencil': 1}
    atoms.calc = GPAW(**params)
    e2 = atoms.get_potential_energy()
    e2a = atoms.calc.get_reference_energy()

    equal(e1a, -2.893 * Hartree, 8e-3)
    equal(e2a, -2.908 * Hartree, 9e-3)
    equal(e1, e2, 4e-3)

    energy_tolerance = 0.0005
    equal(e1, -0.0790449962, energy_tolerance)
    equal(e2, -0.08147563, energy_tolerance)



import numpy as np
from gpaw.utilities.blas import gemmdot
from gpaw.ibz2bz import (get_overlap, get_overlap_coefficients,
                         get_phase_shifted_overlap_coefficients,
                         IBZ2BZMaps)
from gpaw.spinorbit import soc_eigenstates


class Wannier90:
    def __init__(self, calc, seed=None, bands=None, orbitals_ai=None,
                 spin=0, spinors=False):

        if seed is None:
            seed = calc.atoms.get_chemical_formula()
        self.seed = seed

        if bands is None:
            bands = range(calc.get_number_of_bands())
        self.bands = bands

        Na = len(calc.atoms)
        if orbitals_ai is None:
            orbitals_ai = []
            for ia in range(Na):
                ni = 0
                setup = calc.wfs.setups[ia]
                for l, n in zip(setup.l_j, setup.n_j):
                    if not n == -1:
                        ni += 2 * l + 1
                orbitals_ai.append(range(ni))

        self.calc = calc
        self.ibz2bz = IBZ2BZMaps.from_calculator(calc)
        self.bands = bands
        self.Nn = len(bands)
        self.Na = Na
        self.orbitals_ai = orbitals_ai
        self.Nw = np.sum([len(orbitals_ai[ai]) for ai in range(Na)])
        self.kpts_kc = calc.get_ibz_k_points()
        self.Nk = len(self.kpts_kc)
        self.spin = spin
        self.spinors = spinors

        if spinors:
            # spinorbit.WaveFunctions.transform currently do not suppport
            # transformation of wavefunctions, only projections.
            # XXX: should be updated in the future
            assert calc.wfs.kd.nbzkpts == calc.wfs.kd.nibzkpts
            self.soc = soc_eigenstates(calc)
        else:
            self.soc = None

    def write_input(self,
                    mp=None,
                    plot=False,
                    num_iter=100,
                    write_xyz=False,
                    write_rmn=False,
                    translate_home_cell=False,
                    dis_num_iter=200,
                    dis_froz_max=0.1,
                    dis_mix_ratio=0.5,
                    search_shells=None):

        calc = self.calc
        seed = self.seed
        bands = self.bands
        orbitals_ai = self.orbitals_ai
        spinors = self.spinors

        if seed is None:
            seed = calc.atoms.get_chemical_formula()

        if bands is None:
            bands = range(calc.get_number_of_bands())

        Na = len(calc.atoms)
        if orbitals_ai is None:
            orbitals_ai = []
            for ia in range(Na):
                ni = 0
                setup = calc.wfs.setups[ia]
                for l, n in zip(setup.l_j, setup.n_j):
                    if not n == -1:
                        ni += 2 * l + 1
                orbitals_ai.append(range(ni))
        assert len(orbitals_ai) == Na

        Nw = np.sum([len(orbitals_ai[ai]) for ai in range(Na)])
        if spinors:
            Nw *= 2
            new_bands = []
            for n in bands:
                new_bands.append(2 * n)
                new_bands.append(2 * n + 1)
            bands = new_bands

        f = open(seed + '.win', 'w')

        pos_ac = calc.spos_ac
        # pos_av = calc.atoms.get_positions()
        # cell_cv = calc.atoms.get_cell()
        # pos_ac = np.dot(pos_av, np.linalg.inv(cell_cv))

        print('begin projections', file=f)
        for ia, orbitals_i in enumerate(orbitals_ai):
            setup = calc.wfs.setups[ia]
            l_i = []
            n_i = []
            for n, l in zip(setup.n_j, setup.l_j):
                if not n == -1:
                    l_i += (2 * l + 1) * [l]
                    n_i += (2 * l + 1) * [n]
            r_c = pos_ac[ia]
            for orb in orbitals_i:
                l = l_i[orb]
                n = n_i[orb]
                print('f=%1.2f, %1.2f, %1.2f : s ' % (r_c[0], r_c[1], r_c[2]),
                      end='', file=f)
                print('# n = %s, l = %s' % (n, l), file=f)

        print('end projections', file=f)
        print(file=f)

        if spinors:
            print('spinors = True', file=f)
        else:
            print('spinors = False', file=f)
        print('write_hr = True', file=f)
        if write_xyz:
            print('write_xyz = True', file=f)
        if write_rmn:
            print('write_tb = True', file=f)
            print('write_rmn = True', file=f)
        if translate_home_cell:
            print('translate_home_cell = True', file=f)
        print(file=f)
        print('num_bands       = %d' % len(bands), file=f)

        if search_shells is not None:
            print("search_shells = {0}".format(search_shells), file=f)

        maxn = max(bands)
        if maxn + 1 != len(bands):
            diffn = maxn - len(bands)
            print('exclude_bands : ', end='', file=f)
            counter = 0
            for n in range(maxn):
                if n not in bands:
                    counter += 1
                    if counter != diffn + 1:
                        print('%d,' % (n + 1), sep='', end='', file=f)
                    else:
                        print('%d' % (n + 1), file=f)
        print(file=f)

        print('guiding_centres = True', file=f)
        print('num_wann        = %d' % Nw, file=f)
        print('num_iter        = %d' % num_iter, file=f)
        print(file=f)

        if len(bands) > Nw:
            ef = calc.get_fermi_level()
            print('fermi_energy  = %2.3f' % ef, file=f)
            print('dis_froz_max  = %2.3f' % (ef + dis_froz_max), file=f)
            print('dis_num_iter  = %d' % dis_num_iter, file=f)
            print('dis_mix_ratio = %1.1f' % dis_mix_ratio, file=f)
        print(file=f)

        print('begin unit_cell_cart', file=f)
        for cell_c in calc.atoms.cell:
            print('%14.10f %14.10f %14.10f' % (cell_c[0],
                                               cell_c[1],
                                               cell_c[2]),
                  file=f)
        print('end unit_cell_cart', file=f)
        print(file=f)

        print('begin atoms_frac', file=f)
        for atom, pos_c in zip(calc.atoms, pos_ac):
            print(atom.symbol, end='', file=f)
            print('%14.10f %14.10f %14.10f' % (pos_c[0], pos_c[1], pos_c[2]),
                  file=f)
        print('end atoms_frac', file=f)
        print(file=f)

        if plot:
            print('wannier_plot   = True', file=f)
            print('wvfn_formatted = True', file=f)
            print(file=f)

        if mp is not None:
            N_c = mp
        else:
            N_c = calc.wfs.kd.N_c
        print('mp_grid =', N_c[0], N_c[1], N_c[2], file=f)
        print(file=f)
        print('begin kpoints', file=f)

        for kpt in calc.get_bz_k_points():
            print('%14.10f %14.10f %14.10f' % (kpt[0], kpt[1], kpt[2]), file=f)
        print('end kpoints', file=f)

        f.close()

    def write_projections(self):
        calc = self.calc
        seed = self.seed
        spin = self.spin
        orbitals_ai = self.orbitals_ai
        soc = self.soc

        if seed is None:
            seed = calc.atoms.get_chemical_formula()

        bands = get_bands(seed)
        Nn = len(bands)

        spinors = False

        win_file = open(seed + '.win')
        for line in win_file.readlines():
            l_e = line.split()
            if len(l_e) > 0:
                if l_e[0] == 'spinors':
                    spinors = l_e[2]
                    if spinors in ['T', 'true', '1', 'True']:
                        spinors = True
                    else:
                        spinors = False
                if l_e[0] == 'num_wann':
                    Nw = int(l_e[2])
                if l_e[0] == 'mp_grid':
                    Nk = int(l_e[2]) * int(l_e[3]) * int(l_e[4])
                    assert Nk == len(calc.get_bz_k_points())

        Na = len(calc.atoms)
        if orbitals_ai is None:
            orbitals_ai = []
            for ia in range(Na):
                ni = 0
                setup = calc.wfs.setups[ia]
                for l, n in zip(setup.l_j, setup.n_j):
                    if not n == -1:
                        ni += 2 * l + 1
                orbitals_ai.append(range(ni))
        assert len(orbitals_ai) == Na

        if spinors:
            new_orbitals_ai = []
            for orbitals_i in orbitals_ai:
                new_orbitals_i = []
                for i in orbitals_i:
                    new_orbitals_i.append(2 * i)
                    new_orbitals_i.append(2 * i + 1)
                new_orbitals_ai.append(new_orbitals_i)
            orbitals_ai = new_orbitals_ai

        Ni = 0
        for orbitals_i in orbitals_ai:
            Ni += len(orbitals_i)
        assert Nw == Ni

        f = open(seed + '.amn', 'w')

        print('Kohn-Sham input generated from GPAW calculation', file=f)
        print('%10d %6d %6d' % (Nn, Nk, Nw), file=f)

        P_kni = np.zeros((Nk, Nn, Nw), complex)
        for ik in range(Nk):
            if spinors:
                P_ani = soc[ik].P_amj
            else:
                P_ani = get_projections_in_bz(calc.wfs,
                                              ik,
                                              spin,
                                              self.ibz2bz,
                                              bcomm=None)
            for i in range(Nw):
                icount = 0
                for ai in range(Na):
                    ni = len(orbitals_ai[ai])
                    P_ni = P_ani[ai][bands]
                    P_ni = P_ni[:, orbitals_ai[ai]]
                    P_kni[ik, :, icount:ni + icount] = P_ni.conj()
                    icount += ni

        for ik in range(Nk):
            for i in range(Nw):
                for n in range(Nn):
                    P = P_kni[ik, n, i]
                    data = (n + 1, i + 1, ik + 1, P.real, P.imag)
                    print('%4d %4d %4d %18.12f %20.12f' % data, file=f)

        f.close()

    def write_eigenvalues(self):
        calc = self.calc
        seed = self.seed
        spin = self.spin
        soc = self.soc

        bands = get_bands(seed)

        f = open(seed + '.eig', 'w')

        for ik in range(len(calc.get_bz_k_points())):
            if soc is None:
                ibzk = calc.wfs.kd.bz2ibz_k[ik]  # IBZ k-point
                e_n = calc.get_eigenvalues(kpt=ibzk, spin=spin)
            else:
                e_n = soc[ik].eig_m
            for i, n in enumerate(bands):
                data = (i + 1, ik + 1, e_n[n])
                print('%5d %5d %14.6f' % data, file=f)

        f.close()

    def write_overlaps(self, less_memory=False):
        calc = self.calc
        seed = self.seed
        spin = self.spin
        soc = self.soc
        ibz2bz = self.ibz2bz

        if seed is None:
            seed = calc.atoms.get_chemical_formula()

        if soc is None:
            spinors = False
        else:
            spinors = True

        bands = get_bands(seed)
        Nn = len(bands)
        kpts_kc = calc.get_bz_k_points()
        Nk = len(kpts_kc)

        nnkp = open(seed + '.nnkp', 'r')
        lines = nnkp.readlines()
        for il, line in enumerate(lines):
            if len(line.split()) > 1:
                if line.split()[0] == 'begin' and line.split()[1] == 'nnkpts':
                    Nb = eval(lines[il + 1].split()[0])
                    i0 = il + 2
                    break

        f = open(seed + '.mmn', 'w')

        print('Kohn-Sham input generated from GPAW calculation', file=f)
        print('%10d %6d %6d' % (Nn, Nk, Nb), file=f)

        icell_cv = (2 * np.pi) * np.linalg.inv(calc.wfs.gd.cell_cv).T
        r_g = calc.wfs.gd.get_grid_point_coordinates()

        spos_ac = calc.spos_ac
        wfs = calc.wfs
        dO_aii = get_overlap_coefficients(wfs)

        if not less_memory:
            u_knG = []
            for ik in range(Nk):
                u_nG = self.wavefunctions(ik, bands)
                u_knG.append(u_nG)

        proj_k = []
        for ik in range(Nk):
            if spinors:
                proj_k.append(soc[ik].projections)
            else:
                proj_k.append(get_projections_in_bz(calc.wfs,
                                                    ik, spin,
                                                    ibz2bz,
                                                    bcomm=None))

        for ik1 in range(Nk):
            if less_memory:
                u1_nG = self.wavefunctions(ik1, bands)
            else:
                u1_nG = u_knG[ik1]
            for ib in range(Nb):
                # b denotes nearest neighbor k-points
                line = lines[i0 + ik1 * Nb + ib].split()
                ik2 = int(line[1]) - 1
                if less_memory:
                    u2_nG = self.wavefunctions(ik2, bands)
                else:
                    u2_nG = u_knG[ik2]

                G_c = np.array([int(line[i]) for i in range(2, 5)])
                bG_v = np.dot(G_c, icell_cv)
                u2_nG = u2_nG * np.exp(-1.0j * gemmdot(bG_v, r_g, beta=0.0))
                bG_c = kpts_kc[ik2] - kpts_kc[ik1] + G_c
                phase_shifted_dO_aii = get_phase_shifted_overlap_coefficients(
                    dO_aii, spos_ac, -bG_c)
                M_mm = get_overlap(bands,
                                   wfs.gd,
                                   u1_nG,
                                   u2_nG,
                                   proj_k[ik1],
                                   proj_k[ik2],
                                   phase_shifted_dO_aii)
                indices = (ik1 + 1, ik2 + 1, G_c[0], G_c[1], G_c[2])
                print('%3d %3d %4d %3d %3d' % indices, file=f)
                for m1 in range(len(M_mm)):
                    for m2 in range(len(M_mm)):
                        M = M_mm[m2, m1]
                        print('%20.12f %20.12f' % (M.real, M.imag), file=f)

        f.close()

    def write_wavefunctions(self):

        calc = self.calc
        soc = self.soc
        spin = self.spin
        seed = self.seed

        if soc is None:
            spinors = False
        else:
            spinors = True

        if seed is None:
            seed = calc.atoms.get_chemical_formula()

        bands = get_bands(seed)
        Nn = len(bands)
        Nk = len(calc.get_bz_k_points())

        for ik in range(Nk):
            if spinors:
                # For spinors, G denotes spin and grid: G = (s, gx, gy, gz)
                u_nG = soc[ik].wavefunctions(calc, periodic=True)
            else:
                # For non-spinors, G denotes grid: G = (gx, gy, gz)
                u_nG = self.wavefunctions(ik, bands)

            f = open('UNK%s.%d' % (str(ik + 1).zfill(5), spin + 1), 'w')
            grid_v = np.shape(u_nG)[1:]
            print(grid_v[0], grid_v[1], grid_v[2], ik + 1, Nn, file=f)
            for n in range(Nn):
                for iz in range(grid_v[2]):
                    for iy in range(grid_v[1]):
                        for ix in range(grid_v[0]):
                            u = u_nG[n, ix, iy, iz]
                            print(u.real, u.imag, file=f)
            f.close()

    def wavefunctions(self, bz_index, bands):
        if self.spinors:
            # For spinors, G denotes spin and grid: G = (s, gx, gy, gz)
            return self.soc[bz_index].wavefunctions(
                self.calc, periodic=True)[bands]
        # For non-spinors, G denotes grid: G = (gx, gy, gz)
        ibz_index = self.calc.wfs.kd.bz2ibz_k[bz_index]
        ut_nR = np.array([self.calc.wfs.get_wave_function_array(
            n, ibz_index, self.spin,
            periodic=True)
            for n in bands])
        ut_nR_sym = np.array([self.ibz2bz[bz_index].map_pseudo_wave_to_BZ(
            ut_nR[n]) for n in bands])

        return ut_nR_sym


def get_bands(seed):
    win_file = open(seed + '.win')
    exclude_bands = None
    for line in win_file.readlines():
        l_e = line.split()
        if len(l_e) > 0:
            if l_e[0] == 'num_bands':
                Nn = int(l_e[2])
            if l_e[0] == 'exclude_bands':
                exclude_bands = line.split()[2]
                exclude_bands = [int(n) - 1 for n in exclude_bands.split(',')]
    if exclude_bands is None:
        bands = range(Nn)
    else:
        bands = range(Nn + len(exclude_bands))
        bands = [n for n in bands if n not in exclude_bands]
    win_file.close()

    return bands


def get_projections_in_bz(wfs, K, s, ibz2bz, bcomm=None):
    """ Returns projections object in full BZ
    wfs: calc.wfs object
    K: BZ k-point index
    s: spin index
    ibz2bz: IBZ2BZMaps
    bcomm: band communicator
    """
    ik = wfs.kd.bz2ibz_k[K]  # IBZ k-point
    kpt = wfs.kpt_qs[ik][s]
    nbands = wfs.bd.nbands
    # Get projections in ibz
    proj = kpt.projections.new(nbands=nbands, bcomm=bcomm)
    proj.array[:] = kpt.projections.array[:nbands]

    # map projections to bz
    proj_sym = ibz2bz[K].map_projections(proj)
    return proj_sym
